
__all__ = [
    "MolecularHamiltonian"
]

import numpy as np, itertools
import McUtils.Numputils as nput
from McUtils.Combinatorics import UniquePermutations
from McUtils.Zachary import TensorDerivativeConverter

from .CoordinateSystems import MolecularEmbedding, ModeEmbedding
from .Properties import PotentialSurfaceManager, DipoleSurfaceManager, NormalModesManager

class MolecularHamiltonian:
    """
    Provides a generic molecular Hamiltonian with support for expansions (for VPT),
    construction of molecular models, and other wave function generators
    """

    def __init__(self,
                 embedding:MolecularEmbedding,
                 potential_manager:PotentialSurfaceManager=None,
                 modes_manager:NormalModesManager=None,
                 dipole_manager:DipoleSurfaceManager=None
                 ):
        self.embedding = embedding
        self.modes = modes_manager
        self.potential = ScalarOperatorManager(potential_manager)
        self.dipole = ScalarOperatorManager(dipole_manager)

    def get_VPT_expansions(self,
                           order=2,
                           coordinate_transformation=None
                           ):
        carts = self.embedding.internals is None
        embedding = ModeEmbedding(self.embedding, self.modes, dimensionless=True)
        if isinstance(order, (int, np.integer)):
            n = order
            order = {
                'potential':n,
                'kinetic':n,
            }
            if carts:
                order['coriolis'] = n-2
                order['watson'] = n-2
            else:
                order['pseudopotential'] = n-2

        coordinate_transformation = self.prep_transformation(coordinate_transformation)

        expansions = {}
        for k,o in order.items():
            post_processor = None
            if k == 'kinetic':
                generator = self.gmatrix_expansion(embedding=embedding)
            elif k == 'potential':
                # ders = self.potential.derivatives
                # if np.sum(np.abs(ders[0])) > 1e-8: ...
                generator = self.potential_expansion(embedding=embedding, exclude_gradient=True)
                post_processor = lambda exp: exp[1:]
            elif k == 'coriolis':
                generator = self.coriolis_expansion(embedding=embedding)
            elif k == 'pseudopotential':
                generator = self.pseudopotential_expansion(embedding)
            elif k == 'watson':
                generator = self.watson_expansion(embedding)
            else:
                raise ValueError("don't know what to do with key {}".format(k))
            terms = generator.get_terms(o, transformation=coordinate_transformation)
            if post_processor is not None:
                terms = post_processor(terms)
            expansions[k] = terms

        return expansions

    @classmethod
    def prep_transformation(cls, coordinate_transformation):
        if coordinate_transformation is not None:
            if nput.is_numeric(coordinate_transformation[0][0][0]):
                coordinate_transformation = (
                    coordinate_transformation,
                    nput.inverse_transformation(coordinate_transformation, len(coordinate_transformation))
                )

        return coordinate_transformation

    def get_embedding(self, masses=None, modes=True, mass_weight=True, dimensionless=False):
        if modes is True:
            modes = self.modes
        elif modes is False:
            modes = None
        return ModeEmbedding(self.embedding, modes, masses=masses, mass_weight=mass_weight, dimensionless=dimensionless)

    def potential_expansion(self,
                            order=None,
                            *,
                            dimensionless=False,
                            embedding=None,
                            exclude_gradient=False,
                            **embedding_opts
                            ):
        if embedding is None:
            embedding = self.get_embedding(dimensionless=dimensionless, **embedding_opts)

        ders = self.potential.derivatives
        if exclude_gradient:
            ders = [0] + ders[1:]

        exp = ScalarExpansion(embedding, ders)
        if order is not None:
            exp = exp.get_terms(order)
        return exp
    def get_potential_optmizing_transformation(self,
                                               order,
                                               *,
                                               dimensionless=False,
                                               embedding=None,
                                               exclude_gradient=True,
                                               **embedding_opts
                                               ):
        return nput.optimizing_transformation(
            self.potential_expansion(order,
                                     dimensionless=dimensionless,
                                     embedding=embedding,
                                     exclude_gradient=exclude_gradient,
                                     **embedding_opts
                                     ),
            order
        )

    def _get_ke_expansion(self, cls,
                          order=None,
                          *,
                          masses=None,
                          dimensionless=False,
                          embedding=None,
                          **embedding_opts
                          ):
        if embedding is None:
            embedding = self.get_embedding(masses=masses, dimensionless=dimensionless, **embedding_opts)

        exp = cls(embedding)
        if order is not None:
            exp = exp.get_terms(order)
        return exp
    def gmatrix_expansion(self,
                          order=None,
                          *,
                          dimensionless=False,
                          embedding=None,
                          **embedding_opts
                          ) -> "GMatrixExpansion|np.ndarray":
        return self._get_ke_expansion(GMatrixExpansion,
                                      order=order,
                                      dimensionless=dimensionless,
                                      embedding=embedding,
                                      **embedding_opts
                                      )
    def get_kinetic_optmizing_transformation(self,
                                             order,
                                             *,
                                             dimensionless=False,
                                             embedding=None,
                                             **embedding_opts
                                             ):
        g = self.gmatrix_expansion(dimensionless=dimensionless, embedding=embedding, **embedding_opts) #type:GMatrixExpansion
        G_r = g.get_terms(order)

        w = np.diag(G_r[0])
        # Q = [v]
        # R = [v.T]

        # Q0 = Q[0]
        # R0 = R[0]
        # G0 = G_r[0]
        Q = [np.eye(len(w))]
        R = [np.eye(len(w))]

        # w = w
        G_partial = None
        for o in range(1, order + 1):
            G_rem, G_partial = g.reexpress_G(G_r, Q, R,
                                             order=[o],
                                             # GR_expansion=G_partial,
                                             return_GR=True
                                             )
            # G_partial = G_partial[:-1]
            # w = w[np.newaxis]
            ggg = (
                          G_rem[-1] + G_rem[-1].transpose(1, 0, 2)
                   ) / 4
            # print(ggg - ggg.transpose(1, 0, 2))
            # raise ValueError("wtf")
            new_R = -ggg / (w[:, np.newaxis, np.newaxis] + w[np.newaxis, :, np.newaxis])
            # for i in range(o+1):
            #     new_R = np.tensordot(w, new_R)
            # print(w.shape)
            # print(new_R - np.transpose(new_R, (1, 0, 2)))
            # raise Exception(...)
            R = R + [new_R]

            print(new_R - new_R.transpose(1, 0, 2))

            Q = nput.inverse_transformation(R, [o], reverse_expansion=Q)
            print("="*50)
            G_rem2, G_partial = g.reexpress_G(G_r[:1], Q, R,
                                             order=[o],
                                             # GR_expansion=G_partial,
                                             return_GR=True
                                             )
            # print(G_rem[-1])
            print(G_rem2[-1] + G_rem2[-1].transpose(1, 0, 2))

            raise Exception("???")

        return Q, R

    def pseudopotential_expansion(self,
                                  order=None,
                                  *,
                                  dimensionless=False,
                                  embedding=None,
                                  **embedding_opts
                                  ):
        return self._get_ke_expansion(PseudopotentialExpansion,
                                      order=order,
                                      dimensionless=dimensionless,
                                      embedding=embedding,
                                      **embedding_opts
                                      )
    def coriolis_expansion(self,
                           order=None,
                           *,
                           dimensionless=False,
                           embedding=None,
                           **embedding_opts
                           ):
        return self._get_ke_expansion(CoriolisRotationExpansion,
                                      order=order,
                                      dimensionless=dimensionless,
                                      embedding=embedding,
                                      **embedding_opts
                                      )
    def watson_expansion(self,
                         order=None,
                         *,
                         dimensionless=False,
                         embedding=None,
                         **embedding_opts
                         ):
        return self._get_ke_expansion(ReciprocalInertiaExpansion,
                                      order=order,
                                      dimensionless=dimensionless,
                                      embedding=embedding,
                                      **embedding_opts
                                      )

class ScalarOperatorManager:
    def __init__(self, manager_or_derivs):
        self.manager = manager_or_derivs
        self._derivs = None
    @property
    def derivatives(self):
        if hasattr(self.manager, 'derivatives'):
            return self.manager.derivatives
        else:
            return self.manager

class PotentialManager(ScalarOperatorManager):
    def __init__(self, manager_or_derivs, modes:NormalModesManager):
        super().__init__(manager_or_derivs)
        modes = modes.modes
        if modes is not None:
            modes = modes.basis.to_new_modes().make_dimensionless()
        self.modes = modes

    @property
    def derivatives(self):
        if hasattr(self.manager, 'derivatives'):
            derivs = self.manager.derivatives
        else:
            derivs = self.manager
        return self.canonicalize_derivs(derivs)


class ScalarExpansion:
    """
    A helper class that can transform scalar derivatives from Cartesians (or Cartesian modes) to internals
    """
    def __init__(self,
                 embedding:ModeEmbedding,
                 derivs
                 ):
        self.embedding = embedding
        self.derivs = derivs

    def canonicalize_derivs(self, derivs, symmetrize=True):
        if self.embedding.modes is None: return derivs
        in_modes = False
        for d in derivs:
            if not nput.is_zero(d):
                in_modes = d.shape[0] != d.shape[-1]
                if in_modes: break
        YX = self.embedding.mw_inverse()
        if in_modes:
            QX = self.embedding.modes.coords_by_modes @ YX
            Qq = np.diag(1/np.sqrt(self.embedding.modes.freqs))
            _ = []
            for d in derivs:
                if not nput.is_zero(d):
                    for j in range(d.ndim):
                        if d.shape[j] == QX.shape[1]:
                            d = np.tensordot(QX, d, axes=[1, j])
                        else:
                            d = np.tensordot(Qq, d, axes=[1, j])
                        d = np.moveaxis(d, 0, j)
                    # # symmetrize d
                    if symmetrize:
                        n = 0
                        t = 0
                        for p in itertools.permutations(range(d.ndim-1)):
                            n += 1
                            t += d.transpose(p + (d.ndim-1,))
                        d = t / n
                _.append(d)
            derivs = _
        else:
            derivs = nput.tensor_reexpand([YX], derivs, len(derivs))
        return derivs

    def get_terms(self, order=None, *, derivs=None, transformation=None):
        if transformation is not None:
            forward_derivs, reverse_derivs = MolecularHamiltonian.prep_transformation(transformation)
            V = self.get_terms(order, derivs=derivs)
            return nput.tensor_reexpand(forward_derivs, V, order, axes=[-1, 0])

        if derivs is None:
            derivs = self.derivs
        if order is None:
            order = len(self.derivs)
        if len(derivs) < order:
            derivs = tuple(derivs) + (0,) * (order - len(derivs))

        derivs = self.canonicalize_derivs(derivs)

        zero_derivs = 0
        for d in derivs:
            if isinstance(d, (int, np.integer, float, np.floating)) and d == 0:
                d+=1
            else:
                break

        QX = self.embedding.get_cartesians_by_internals(order=order - zero_derivs)
        real_ders = [d for d in derivs if not nput.is_zero(d)]
        if self.embedding.modes is not None and (
                real_ders[0].shape[0] == self.embedding.modes.modes_by_coords.shape[1]
        ):
            XQ = self.embedding.modes.modes_by_coords
            QX = nput.tensor_reexpand(QX, [XQ], order - zero_derivs, axes=[-1, 0])
        else:
            YX = self.embedding.mw_inverse()
            QX = [np.tensordot(q, YX, axes=[-1, 0]) for q in QX]

        # return TensorDerivativeConverter(QX, derivs).convert()
        return nput.tensor_reexpand(QX, derivs, order, axes=[-1, 0])

class GMatrixExpansion:
    # I thought about having this be a generic "tensor product" expansion, but this is cleaner
    def __init__(self, embedding:ModeEmbedding):
        self.embedding = embedding

    def get_terms(self, order, transformation=None):
        if transformation is not None:
            forward_derivs, reverse_derivs = MolecularHamiltonian.prep_transformation(transformation)
            return self.reexpress(order, forward_derivs, reverse_derivs)
        XQ = self.embedding.get_internals_by_cartesians(order=order+1)
        XG = nput.tensordot_deriv(XQ, XQ, order, axes=[0, 0])#, identical=False)

        if order > 0:
            QX = self.embedding.get_cartesians_by_internals(order=order+1)
            # XG = self.reexpress_G(XG, QX, XQ, order=order)
            XG = [XG[0]] + nput.tensor_reexpand(QX, XG[1:], order)
        return XG

    @classmethod
    def _dRGQ_partition_contrib(cls, partition, R, G):
        print(partition)
        r1, r2, s = partition
        if s >= len(G): return 0
        if r1 >= len(R) or r2 >= len(R): return 0

        base_term = G[s]
        if isinstance(base_term, (int, float, np.integer, np.floating)) and base_term == 0:
            return 0

        r_perm_counter = 1
        r_perm_idx = []
        # g_perm_idx = []

        a = base_term.ndim - 2
        for r in [r1, r2]:
            d = R[r]
            if nput.is_zero(d): return 0

            if r == 0:
                base_term = np.moveaxis(base_term, a, -1)
            else:
                base_term = np.tensordot(base_term, d, axes=[a, 0])
                r_perm_idx.extend([r_perm_counter] * r)
                r_perm_counter -= 1
        if r2 > 0:
            # 1 1 -> (x, i, y, j) -> (-3 -> -2)
            base_term = np.moveaxis(base_term, -(r2 + 2), -2) # axis from the r1 contraction
        # if s > 0:
        g_perm_idx = ([1]*s) + ([0]*(r1+r2))

        if r1 > 0 or r2 > 0:
            # r1 and r2 need to be permuted to preserve symmetry, but all perms can overcount
            nterms = TensorDerivativeConverter.compute_partition_terms([p for p in partition[:2] if p > 0])
            r_perms, _ = UniquePermutations(r_perm_idx).permutations(return_indices=True)

            overcount = len(r_perms) / nterms
            base_term = base_term / overcount

            # if r1 != r2:
            base_term = base_term + np.moveaxis(base_term, -1, -2)


            # g indices need to be swapped into the r coords sometimes too
            g_perms, _ = UniquePermutations(g_perm_idx).permutations(return_indices=True)

            # take direct product of permutation indices
            perm_inds = []
            # print(overcount)
            padding = list(range(base_term.ndim - 2, base_term.ndim))
            for p in g_perms:
                for r in r_perms:
                    perm_inds.append(
                        list(p[:s]) + list(p[s:][r]) + padding
                    )

            base_term = sum(
                base_term.transpose(p)
                for p in perm_inds
            )

        return base_term

    @classmethod
    def _dRGQ_derivs(cls, R, G, o):
        # parts = IntegerPartitionPermutations(o, dim=3).get_partition_permutations(flatten=True)

        # print("="*50)
        total_cont = 0
        for g in range(o+1):
            rem = (o+1)-g
            # don't want to go smaller than half of rem, b.c. symmetry
            for r1 in range(rem//2, rem):
                r2 = o - (g + r1)
                # print("-"*10)
                # print(r1, r2, g)
                pc = cls._dRGQ_partition_contrib([r1, r2, g], R, G)
                # print(pc[0])
                total_cont += pc
        # print("_"*50)
        return total_cont

    @classmethod
    def reexpress_G(self,
                    G_expansion, forward_derivs, reverse_derivs,
                    order,
                    *,
                    GR_expansion=None,
                    return_GR=False
                    ):
        """
        Apply a coordinate transformation to the G-matrix

        :param forward_derivs:
        :param reverse_derivs:
        :param order:
        :return:
        """
        R = reverse_derivs
        Q = forward_derivs
        G = G_expansion

        if nput.is_numeric(order): order = list(range(order+1))
        m = max(order)

        G_R = GR_expansion
        if G_R is None:
            G_R = [self._dRGQ_derivs(R, G, o) for o in range(1, m+1)]

        if len(G_R) < m:
            G_R = G_R + [self._dRGQ_derivs(R, G, o) for o in range(len(G_R)+1, m+1)]

        rexpansion = nput.tensor_reexpand(Q,
                                          G_R,
                                          order=[o for o in order if o > 0],
                                          axes=[-1, 0]
                                          )
        G_Q = []
        n = 0
        G0 = None
        for o in order:
            if o == 0:
                if G0 is None:
                    G0 = reverse_derivs[0] @ G[0] @ reverse_derivs[0].T
                G_Q.append(G0)
            else:
                G_Q.append(rexpansion[n])
                n += 1

        if return_GR:
            return G_Q, G_R
        else:
            return G_Q

    def reexpress(self, order, forward_derivs, reverse_derivs):
        return self.reexpress_G(
            self.get_terms(order),
            forward_derivs,
            reverse_derivs,
            order=order
        )

class PseudopotentialExpansion:
    def __init__(self, g_expansion:"ModeEmbedding|GMatrixExpansion"):
        if not isinstance(g_expansion, GMatrixExpansion):
            g_expansion = GMatrixExpansion(g_expansion)
        self.g_expansion = g_expansion

    @classmethod
    def lambda_expansion(cls,
                        G_expansion, G_inv,
                        I0_expansion, I0_inv,
                        order):

        g_exp = [np.trace(t, axis1=-1, axis2=-2) for t in nput.tensordot_deriv(G_expansion[1:], G_inv, order)]
        i_exp = [np.trace(t, axis1=-1, axis2=-2) for t in nput.tensordot_deriv(I0_expansion[1:], I0_inv, order)]
        return [ti - tg for ti,tg in zip(i_exp, g_exp)]

    @classmethod
    def get_U(cls, G_expansion, I0_expansion):

        QG = G_expansion
        QI0 = I0_expansion
        n = len(QG) - 1
        order = n - 2

        QU = nput.matinv_deriv(QI0, n)
        QiG = nput.matinv_deriv(G_expansion, n)

        L = cls.lambda_expansion(
            QG, QiG,
            QI0, QU,
            order + 1
        )

        LxL = nput.tensorprod_deriv(L, L, order)

        inner_exp = [
            l1 + 1/4*l2
            for l1, l2 in zip(L[1:], LxL)
        ]
        G_contract = nput.tensordot_deriv(G_expansion, inner_exp, order, axes=[[0, 1], [0, 1]])
        QG_contract = [
            np.trace(g, axis1=-1, axis2=-2) for g in
            nput.tensordot_deriv(G_expansion[1:], L, order, axes=[2, 0])
        ]

        return [g1 + g2 for g1, g2 in zip(G_contract, QG_contract)]
        # return nput.scalarprod_deriv(Qig, t, order=order)

    def get_terms(self, order, transformation=None):
        if transformation is not None:
            forward_derivs, reverse_derivs = MolecularHamiltonian.prep_transformation(transformation)
            return self.reexpress(order, forward_derivs, reverse_derivs)

        # build setup terms, need to go 2 orders past the nominal order
        n = order + 2
        QG = self.g_expansion.get_terms(n)
        QI0 = self.g_expansion.embedding.get_inertia_tensor_expansion(order=n)

        return self.get_U(QG, QI0)

    def reexpress(self, order, forward_derivs, reverse_derivs):
        """
        Apply a coordinate transformation to the G-matrix

        :param forward_derivs:
        :param reverse_derivs:
        :param order:
        :return:
        """

        # build setup terms, need to go 2 orders past the nominal order
        n = order + 2
        QG = self.g_expansion.reexpress(n, forward_derivs, reverse_derivs)
        QI0 = nput.tensor_reexpand(
            forward_derivs,
            self.g_expansion.embedding.get_inertia_tensor_expansion(order=n),
            n
        )

        return self.get_U(QG, QI0)

class ZetaExpansion:
    def __init__(self, embedding:ModeEmbedding):
        self.embedding = embedding
    def get_terms(self, order, transformation=None):
        QY = self.embedding.get_cartesians_by_internals(order=order)
        if transformation is not None:
            forward_derivs, reverse_derivs = MolecularHamiltonian.prep_transformation(transformation)
            QY = nput.tensor_reexpand(forward_derivs, QY, order)

        B_e, eigs = self.embedding.get_inertial_frame()

        _ = []
        for J in QY:
            J = J.reshape(J.shape[:-2] + (-1, 3))
            J = np.tensordot(eigs, J, axes=[1, 2])
            _.append(np.moveaxis(J, -2, -1))
        QY = _

        QYQY = [
            np.sum(qyqy, axis=-1)
            for qyqy in nput.tensorprod_deriv(QY, QY, axes=[[0, 1], [0, 1]], order=order)
        ]

        zeta_expansion = []
        for qyqy in QYQY:
            for a in range(3):
                b = (a + 1) % 3
                c = (a + 2) % 3
                if b % 2 == 1:
                    c, b = b, c
                zeta_expansion.append(qyqy[..., b, :, c] - qyqy[..., c, :, b])

        return zeta_expansion

class ReciprocalInertiaExpansion:
    """
    Pulled from Watson
    """
    def __init__(self, embedding:ModeEmbedding):
        self.embedding = embedding

    def get_terms(self, order, transformation=None):
        I0, a = self.embedding.get_inertia_tensor_expansion(order=1)
        if transformation is not None:
            forward_derivs, reverse_derivs = MolecularHamiltonian.prep_transformation(transformation)
            a = nput.tensor_reexpand(forward_derivs, [a], 1)[0]
        u = np.linalg.inv(I0)

        expansion = [u]
        for _ in range(order):
            prev = expansion[-1]
            new = np.tensordot(
                np.tensordot(a, prev, axes=[-1, -1]),
                u,
                axes=[-1, 0]
            )
            expansion.append(new)

        return expansion

class CoriolisRotationExpansion:
    """
    Provides an expansion of the Coriolis rotation operator
    """
    def __init__(self, embedding:ModeEmbedding):
        self.embedding = embedding
        self.zetas = ZetaExpansion(self.embedding)
        self.us = ReciprocalInertiaExpansion(self.embedding)

    def get_terms(self, order, transformation=None):
        Z = self.zetas.get_terms(order, transformation=transformation)
        ZZ = nput.tensorprod_deriv(Z, Z, order, axes=[1, 1]) # x/y/z shared
        u = self.us.get_terms(order, transformation=transformation)

        return nput.tensordot_deriv(u, ZZ, order, axes=[1, 0])



"""
An implementation of vibrational perturbation theory (VPT) that uses sparse matrix methods to obtain
corrections.
Makes heavy use of the `BasisReps` package as well as `McUtils.Coordinerds` to obtain representations
of the corrections to the vibrational Hamiltonian.
Is technically not restricted to VPT in a harmonic basis, but no other forms of PT are likely to
be implemented in the near future.
For the purposes of papers, we've been calling this implementation `PyVibPTn`.

The easiest way to run jobs is through the `VPTRunner.run_simple` interface.
The options for jobs along with short descriptions are detailed in
[`VPTSystem`](VPT2/Runner/VPTSystem.md) for molecule/system-related options,
[`VPTStateSpace`](VPT2/Runner/VPTStateSpace.md) for state-space & degeneracy-related options,
[`VPTHamiltonianOptions`](VPT2/Runner/VPTHamiltonianOptions.md) for expansion-related options,
[`VPTHamiltonianOptions`](VPT2/Runner/VPTHamiltonianOptions.md) for expansion-related options,
[`VPTSolverOptions`](VPT2/Runner/VPTSolverOptions.md) for options related to constructing representations and applying VPT,
and [`VPTRuntimeOptions`](VPT2/Runner/VPTRuntimeOptions.md) for options related to the runtime/logging

**A basic tutorial to provide more extensive hand-holding can be found [here](VPT2/tutorial.md).**

Finally, the general code flow is detailed below

![pt design](/Psience/img/PyVibPTnDesign.png){:width="100%"}

:long_description: The implementation of vibrational perturbation theory provided here uses a kernel/config/driver type of design.
The kernel that actually solves the perturbation theory equations is the [`PerturbationTheorySolver`](PerturbationTheorySolver.md) object.
The config comes the [`PerturbationTheoryHamiltonian`](PerturbationTheoryHamiltonian.md), which implements the expansion of the Hamiltonian
with respect to normal modes.
Finally, the driver is the [`VPTRunner`](VPTRunner.md) which through its `run_simple` method aggregates all of the possible options needed
for VPT and sends them to the right parts of the architecture.

Because of this, while it can be helpful to know how the `PerturbationTheorySolver` and `PerturbationTheoryHamiltonian` work, all that one
usually needs to do to run VPT is to call `VPTRunner.run_simple`.
The most basic call looks like
```python
VPTRunner.run_simple(system_spec, states)
```
where the `system_spec` is often an `fchk` file from an electronic structure calculation that provides a restricted quartic potential
and the `states` is an `int` that specifies the max number of quanta of excitation in the states that will be corrected.

The system spec can also be more explicit, being either a `Molecule` object or a list like `[atoms, coords, opts]`
where `atoms` is a list of strings of the atoms, `coords` are the accompanying Cartesian coordinates, and `opts` is a `dict`
of options for the molecule, such as the `masses`.
It is also possible (and sometimes necessary) to supply custom normal modes, which can be done through the `modes` option (examples below).
"""

__all__ = []
from .Runner import *; from .Runner import __all__ as exposed
__all__ += exposed
from .Analyzer import *; from .Analyzer import __all__ as exposed
__all__ += exposed
from .Hamiltonian import *; from .Hamiltonian import __all__ as exposed
__all__ += exposed
from .Solver import *; from .Solver import __all__ as exposed
__all__ += exposed
from .Corrections import *; from .Corrections import __all__ as exposed
__all__ += exposed
from .Wavefunctions import *; from .Wavefunctions import __all__ as exposed
__all__ += exposed
from .Terms import *; from .Terms import __all__ as exposed
__all__ += exposed

from .Analytic import *; from .Analytic import __all__ as exposed
__all__ += exposed


"""
Provides classes to support wave functions coming from VPT calculations
"""

import numpy as np, itertools as ip, time, enum, math

from McUtils.Numputils import SparseArray
import McUtils.Numputils as nput
from McUtils.Scaffolding import ParameterManager, Checkpointer, NullCheckpointer
from McUtils.Data import UnitsData

from ..Wavefun import Wavefunctions
from ..BasisReps import SimpleProductBasis, BraKetSpace

from .Terms import DipoleTerms, OperatorTerms
from .Solver import PerturbationTheoryCorrections

__all__ = [
    # 'PerturbationTheoryWavefunction',
    'PerturbationTheoryWavefunctions'
]

__reload_hook__ = ["..Wavefun"]


# class PerturbationTheoryWavefunction(ExpansionWavefunction):
#     """
#     These things are fed the first and second order corrections
#     """
#
#     def __init__(self, mol, basis, corrections):
#         """
#         :param mol: the molecule the wavefunction is for
#         :type mol: Molecule
#         :param basis: the basis the expansion is being done in
#         :type basis: RepresentationBasis
#         :param corrections: the corrections to the terms
#         :type corrections: PerturbationTheoryCorrections
#         """
#         self.mol = mol
#         self.corrs = corrections
#         self.rep_basis = basis
#         super().__init__(self.corrs.energies, self.corrs.wavefunctions, None)
#
#     @property
#     def order(self):
#         return self.corrs.order
#
#     def expectation(self, operator, other):
#         return NotImplemented
#
#     @property
#     def zero_order_energy(self):
#         return self.corrs.energies[0]

class PerturbationTheoryWavefunctions(Wavefunctions):
    """
    These things are fed the first and second order corrections
    """

    def __init__(self,
                 mol, basis, corrections,
                 initial_states=None,
                 modes=None,
                 mode_selection=None,
                 logger=None,
                 checkpoint=None,
                 results=None,
                 operator_settings=None,
                 expansion_options=None,
                 degenerate_transformation_layout=None # need to see if this is really necessary...
                 ):
        """
        :param mol: the molecule the wavefunction is for
        :type mol: Molecule
        :param basis: the basis the expansion is being done in
        :type basis: SimpleProductBasis
        :param corrections: the corrections to the terms
        :type corrections: PerturbationTheoryCorrections
        """
        super().__init__(None, corrections) # just to

        self.mol = mol
        self.corrs = corrections
        self.modes = modes
        self.mode_selection = mode_selection
        self.rep_basis = basis  # temporary hack until I decided how to merge the idea of
        # AnalyticWavefunctions with a RepresentationBasis
        self.initial_states = initial_states # For tracking where to start when calculating properties
        self._init_inds = None
        self._tm_dat = None
        self._dipole_terms = None
        self._dipole_partitioning = self.DipolePartitioningMethod.Standard
        self.logger = logger

        self.checkpointer = Checkpointer.build_canonical(checkpoint)
        if results is None:
            self.results = self.checkpointer
        else:
            self.results = Checkpointer.build_canonical(results)
            
        if expansion_options is None:
            expansion_options = {}
        self.expansion_options = expansion_options
        self.operator_settings = operator_settings if operator_settings is not None else {}
        if 'skipped_coefficient_threshold' in self.operator_settings:
            self.operator_settings = self.operator_settings.copy()
            del self.operator_settings['skipped_coefficient_threshold']
        self._order = None
        self.degenerate_transformation_layout="column" if degenerate_transformation_layout is None else degenerate_transformation_layout

        # super().__init__(
        #     self.corrs.energies,
        #     self.corrs.wfn_corrections,
        #     None
        # )

    def get_dimension(self):
        return len(self.modes)

    @property
    def energies(self):
        return self.corrs.energies
    @energies.setter
    def energies(self, engs):
        if engs is not None:
            raise ValueError("can't set energies")

    def to_state(self, serializer=None):
        keys = dict(
            corrections=self.corrs,
            molecule=self.mol,
            modes=self.modes,
            mode_selection=self.mode_selection,
            basis=self.rep_basis,
            expansion_options=self.expansion_options,
            operator_settings=self.operator_settings,
            logger=self.logger
        )
        return keys

    @classmethod
    def from_state(cls, data, serializer=None):
        return cls(
            serializer.deserialize(data['molecule']),
            serializer.deserialize(data['basis']), serializer.deserialize(data['corrections']),
            modes=serializer.deserialize(data['modes']),
            mode_selection=serializer.deserialize(data['mode_selection']),
            logger=serializer.deserialize(data['logger']),
            operator_settings=serializer.deserialize(data['operator_settings']),
            expansion_options=serializer.deserialize(data['expansion_options'])
        )

    @property
    def degenerate_transformation(self):
        return self.corrs.degenerate_transf

    @property
    def initial_state_indices(self):
        if self._init_inds is None:
            if self.initial_states is not None:
                self._init_inds = self.corrs.states.find(self.initial_states)
            else:
                self._init_inds = (0,)
        return self._init_inds

    def energies_to_order(self, order=None):
        if order is None:
            order = self.corrs.order
        return np.sum(self.corrs.energy_corrs[:, :order+1], axis=1)
    @property
    def deperturbed_energies(self):
        return self.energies_to_order()
    def deperturbed_frequencies(self, order=None):
        if self.initial_states is not None:
            raise NotImplementedError("haven't initial_spaces to this yet")
        eng = self.energies_to_order(order=order)
        return eng[1:] - eng[0]

    @property
    def order(self):
        if self._order is None:
            return self.corrs.order
        else:
            return self._order

    def expectation(self, operator, other=None):
        raise NotImplementedError("general expectation values not implemented yet...")

    @property
    def zero_order_energies(self):
        return self.corrs.energy_corrs[:, 0]

    def get_M0(self, mu_0):
        return self.rep_basis.representation(name='M(0)', coeffs=mu_0,
                                             axes=[[], []],
                                             **self.operator_settings
                                             )
    def get_M1(self, mu_1):
        return self.rep_basis.representation("x", name='M(1)', coeffs=mu_1,
                                             axes=[[0], [1]],
                                             **self.operator_settings
                                             )
    def get_M2(self, mu_2):
        return 1 / 2 * self.rep_basis.representation("x", "x", name="M(2)",
                                                     coeffs=mu_2,
                                                     axes=[[0, 1], [1, 2]],
                                                     **self.operator_settings
                                                     )
    def get_M3(self, mu_3):
        return 1 / 6 * self.rep_basis.representation("x", "x", "x", name="M(3)", coeffs=mu_3,
                                                     axes=[[0, 1, 2], [1, 2, 3]],
                                                     **self.operator_settings
                                                     )
    def get_Mi(self, i, mu, base_sym="M"):
        extra = mu.ndim - i
        return 1 / math.factorial(i) * self.rep_basis.representation(
            *(("x",) * i), name="{}({})".format(base_sym, i), coeffs=mu,
            axes=[list(range(0, i)), list(range(extra, i + extra))],
            **self.operator_settings
        )

    # @profile
    def _get_rep_states(self, m, k, order, ket_spaces, bra_spaces):
            with self.logger.block(tag="getting coupled states for {}".format(m)):
                start = time.time()
                inds = None
                rules = m.selection_rules
                # bra_space = bra_spaces.to_single().take_unique()
                # ket_space = ket_spaces.to_single().take_unique()
                # inds, filter = bra_space.get_representation_brakets(
                #     other=ket_space, return_filter=True,
                #     selection_rules=m.selection_rules
                # )

                for i in range(order):
                    ket_space = None
                    for j in range(order):
                        if i + j + k <= (order-1):
                            # self.logger.log_print("<{i}|{k}|{j}>", i=i, j=j, k=k)
                            for nk, kets in enumerate(ket_spaces.spaces):
                                if ket_space is None:
                                    ket_space = kets[j]
                                else:
                                    ket_space = ket_space.union(kets[j])

                    bra_space = None
                    for nb, bras in enumerate(bra_spaces.spaces):
                        if ket_space is None:
                            bra_space = bras[i]
                        else:
                            bra_space = ket_space.union(bras[i])

                    # self.logger.log_print("<{bras}|{k}|{kets}>", bras=bra_space, kets=ket_space, k=k)

                    if ket_space is not None:
                        filt=None
                        new_inds, filt = bra_space.get_representation_brakets(
                            other=ket_space, filter=filt, return_filter=True,
                            selection_rules=rules
                        )  # no selection rules here
                        # filters[nk][j] = filter
                        if inds is None:
                            inds = new_inds #type: BraKetSpace
                        else:
                            inds = inds.concatenate(new_inds)
                end = time.time()

                self.logger.log_print([
                    'got {sp}',
                    'took {t:.3f}s...'
                    ],
                    sp=inds,
                    t=end - start
                )

            if inds is None:
                raise ValueError("Dipole operator {} (order {}) has empty representation at order {}".format(
                    m,
                    k,
                    order-1
                ))

            inds = inds.remove_duplicates(assume_symmetric=True)

            return inds

    @staticmethod
    def _is_zero(x):
        return isinstance(x, (int, float, np.integer, np.floating)) and x == 0
    # from memory_profiler import profile
    # @profile
    def _mu_representations(self,
                            mu,
                            M,
                            space,
                            bra_spaces,
                            ket_spaces,
                            order,
                            partitioning,
                            rep_inds,
                            allow_higher_dipole_terms=False
                            ):

        # define out reps based on partitioning style
        if partitioning is None:
            partitioning = self.dipole_partitioning
        elif not isinstance(partitioning, self.DipolePartitioningMethod):
            partitioning = self.DipolePartitioningMethod(partitioning)

        # filters = [[None] * (order+1) for _ in range(ket_spaces.nstates)]
        def get_states(m, k):
            return self._get_rep_states(m, k, order, ket_spaces, bra_spaces)

        if all(
                    isinstance(m, (np.ndarray, SparseArray)) and m.shape == (M, M)
                    or self._is_zero(m)
                    for m in mu
        ): # we were passed representations to reuse
            mu_terms = mu
            if (
                    partitioning == self.DipolePartitioningMethod.Standard and
                    order > 1
                    and len(mu_terms) < 4
            ):
                mu_3 = [d[3] for d in self.dipole_terms]
                m3 = self.get_M3(mu_3)
                if rep_inds[3] is None:
                    m3_inds = get_states(m3, 2)
                    with self.logger.block(tag="building {h}".format(h=m3)):
                        rep_inds[3] = m3_inds
                        rep_3 = m3.get_representation_matrix(rep_inds[3], self.corrs.total_basis, remove_duplicates=False)
                mu_terms.append(rep_3)
                # mu_terms = mu_terms + [rep_3]
        else:
            int_part = partitioning == self.DipolePartitioningMethod.Intuitive


            mu_coeffs = []
            # mu_0 = np.array([m[0] for m in mu])
            # mu_coeffs.append(mu_0)
            # if (not int_part and order > 0) or (int_part and order > 1):
            #     mu_1 = np.array([m[1] for m in mu])
            # else:
            #     mu_1 = None
            # mu_coeffs.append(mu_1)
            # if (not int_part and order > 1) or (int_part and order > 2):
            #     mu_2 = np.array([m[2] for m in mu])
            # else:
            #     mu_2 = None
            # mu_coeffs.append(mu_2)
            # if (not int_part and order > 2) or (int_part and order > 3):
            #     mu_3 = np.array([m[3] for m in mu])
            # else:
            #     mu_3 = None
            # mu_coeffs.append(mu_3)

            n = self.corrs.states.ndim
            for i in range(order + (1 if not int_part else 0)):
                mu_sel = []
                if self._is_zero(mu):
                    mu_sel = [np.zeros((n,)*i), np.zeros((n,)*i), np.zeros((n,)*i)]
                else:
                    for m in mu:
                        if self._is_zero(m):
                            mu_sel.append(np.zeros((n,)*i))
                        else:
                            try:
                                mu_sel.append(m[i])
                            except IndexError:
                                if allow_higher_dipole_terms:
                                    mu_sel.append(np.zeros((n,)*i))
                                else:
                                    raise ValueError("dipole expansion not provided to order {}".format(i))

                mu_coeffs.append(np.array(mu_sel))

            # if (not int_part and order > 3) or (int_part and order > 4):
            #     raise NotImplementedError("dipole representations up to order {} with {} partitioning not supported yet".format(
            #         order,
            #         self.DipolePartitioningMethod
            #     ))

            reps = []
            for i, mu in enumerate(mu_coeffs):
                m = self.get_Mi(i, mu)
                if partitioning == self.DipolePartitioningMethod.Intuitive:
                    inds = get_states(m, i)
                elif partitioning == self.DipolePartitioningMethod.Standard:
                    inds = get_states(m, max(0, i-1))
                else:
                    raise ValueError("don't know how to interpret dipole partitioning {}".format(partitioning))

                if i >= len(rep_inds):
                    for d in range(i-len(rep_inds) + 1):
                        rep_inds.append(None)
                rep_inds[i] = inds
                reps.append(m)

            # reps = [m1, m2, m3]
            mu_terms = [None] * len(reps)
            tb = self.corrs.total_basis
            # with self.corrs.disk_backed():

            for i, h in enumerate(reps):
                m_pairs = rep_inds[i]
                if m_pairs is None:
                    raise ValueError("representation indices haven't filled enough to calculate {}".format(h))

                with self.logger.block(tag="building {}".format(h)):
                    start = time.time()
                    sub = h.get_representation_matrix(m_pairs, tb
                                                      # TODO: clearly there's some kind of issue in M(3)?
                                                      , zero_element_warning=False # expect zeros in M(3)...?
                                                      # , remove_duplicates=False
                                                      )
                    end = time.time()
                    self.logger.log_print('took {t:.3f}s...', t=end - start)
                # sub = generate_rep(h, m_pairs)
                mu_terms[i] = sub

        return mu_terms

    # @profile
    def _transition_moments(self,
                            mu_x, mu_y, mu_z,
                            correction_terms=None,
                            lower_states=None,
                            excited_states=None,
                            degenerate_transformation=None,
                            partitioning=None,
                            order=None
                            ):
        """
        Calculates the x, y, and z components of the transition moment between
        the wavefunctions stored.
        By default, only calculates moments involving the ground states.

        :param mu_x: dipole x components (1st, 2nd, 3rd derivatives in normal modes)
        :type mu_x: Iterable[np.ndarray]
        :param mu_y: dipole y components (1st, 2nd, 3rd derivatives in normal modes)
        :type mu_y: Iterable[np.ndarray]
        :param mu_z: dipole z components (1st, 2nd, 3rd derivatives in normal modes)
        :type mu_z: Iterable[np.ndarray]
        :param partitioning: whether to put linear and constant in mu_0 (standard) or linear in mu_1 (intuitive)
        :type partitioning: None | str
        :return:
        :rtype:
        """

        if order is None:
            exp_order = None if 'expansion_order' not in self.expansion_options else self.expansion_options['expansion_order']
            if exp_order is None:
                exp_order = self.order
            elif isinstance(exp_order, int):
                exp_order = exp_order + 1
            else:
                if 'dipole' in exp_order:
                    exp_order = exp_order['dipole'] + 1
                elif 'default' in exp_order:
                    exp_order = exp_order['default'] + 1
                else:
                    exp_order = self.order
            order = min([exp_order, self.order])

        logger = self.logger
        with logger.block(tag="Calculating intensities:", printoptions={'linewidth':int(1e8)}):

            space = self.corrs.coupled_states
            # store this for handling the degenerate_transformation
            lower_states_input = self.initial_state_indices if lower_states is None else lower_states
            upper_states_input = tuple(range(space.nstates)) if excited_states is None else excited_states

            if lower_states is None:
                low_spec = self.initial_state_indices
            else:
                if isinstance(lower_states, (int, np.integer)):
                    lower_states = (lower_states,)
                low_spec = lower_states
            lower_states = space[(low_spec,)]

            if excited_states is None:
                up_spec = tuple(range(space.nstates))
                excited_states = space
            else:
                if isinstance(excited_states, (int, np.integer)):
                    excited_states = (excited_states,)
                up_spec = excited_states
                excited_states = space[(up_spec,)]

            bra_spaces = lower_states  # if isinstance(lower_states, BasisStateSpace) else lower_states.to_single().take_unique()
            ket_spaces = excited_states  # if isinstance(excited_states, BasisStateSpace) else excited_states.to_single().take_unique()

            # M = len(space.indices)
            logger.log_print(
                [
                    "lower/upper states: {l}/{u}"
                ],
                l=len(low_spec),
                u=len(up_spec)
            )

            if correction_terms is None:
                with logger.block(tag='taking correction subspaces:'):
                    start = time.time()
                    logger.log_print('getting ground space...')
                    corr_terms_lower = [self.corrs.wfn_corrections[i][low_spec, :] for i in range(order)]
                    logger.log_print(
                        [
                            "highest-order space: {s}",
                        ],
                        s=corr_terms_lower[-1]
                    )

                    logger.log_print('getting excited space...')
                    corr_terms_upper = [self.corrs.wfn_corrections[i][up_spec, :].T for i in range(order)]
                    logger.log_print(
                        [
                            "highest-order space: {s}",
                        ],
                        s=corr_terms_upper[-1]
                    )

                    end = time.time()
                    logger.log_print(
                        [
                            "took {t:.3f}s..."
                        ],

                        t=end - start
                    )
            else:
                corr_terms_lower, corr_terms_upper = correction_terms

            rep_inds = [None] * 4
            M = self.corrs.wfn_corrections[0].shape[1]
            total_space = self.corrs.total_basis
            transition_moment_components = np.zeros((order, 3)).tolist()  # x, y, and z components of the 0th, 1st, and 2nd order stuff

            mu = [mu_x, mu_y, mu_z]

            # define out reps based on partitioning style
            if partitioning is None:
                partitioning = self.dipole_partitioning
            elif not isinstance(partitioning, self.DipolePartitioningMethod):
                partitioning = self.DipolePartitioningMethod(partitioning)

            logger.log_print(
                [
                    "dipole partitioning: {m}"
                ],
                m=partitioning.value
            )

            # raise Exception(corr_terms[1], corr_terms_lower[1], corr_terms_upper[1], low_spec, len(low_spec), len(up_spec))

            mu_reps = self._mu_representations(
                mu,
                M,
                total_space,
                bra_spaces,
                ket_spaces,
                order,
                partitioning,
                rep_inds,
                allow_higher_dipole_terms=self.expansion_options.get('allow_higher_dipole_terms', True)
            )

            # raise Exception("...")
                # logger.log_print(
                #     [
                #         "took {t:.3f}s..."
                #     ],
                #     t=end-start
                # )

            # raise Exception(mu_terms)
            if degenerate_transformation is None:
                degenerate_transformation = self.corrs.degenerate_transf
            if degenerate_transformation is not None:
                    # raise  Exception("ugh")
                    # degenerate_transformation = degenerate_transformation.T

                if len(lower_states_input) > 1:
                    deg_transf_lower = degenerate_transformation[np.ix_(lower_states_input, lower_states_input)]
                    if self.degenerate_transformation_layout == "column":
                        deg_transf_lower = deg_transf_lower.T
                else:
                    deg_transf_lower = None
                if len(upper_states_input) > 1:
                    if self.degenerate_transformation_layout == "column":
                        deg_transf_upper = degenerate_transformation[:, upper_states_input]
                    else:
                        deg_transf_upper = degenerate_transformation[upper_states_input, :].T
                else:
                    deg_transf_upper = None
            else:
                deg_transf_lower = deg_transf_upper = None

            if degenerate_transformation is not None:
                transition_moment_components_deg = np.zeros((order, 3)).tolist()  # x, y, and z components of the 0th, 1st, and 2nd order stuff
            else:
                transition_moment_components_deg = None

            corr_terms = self.corrs.wfn_corrections
            for a in range(3):
                mu_basic_terms = [r[:, :, a] for r in mu_reps]
                if partitioning == self.DipolePartitioningMethod.Intuitive:
                    mu_terms = [mu_basic_terms[0]]
                    if order > 1:
                        mu_terms.append(mu_basic_terms[1])
                    if order > 2:
                        mu_terms.append(mu_basic_terms[2])
                elif partitioning == self.DipolePartitioningMethod.Standard:
                    mu_terms = [mu_basic_terms[0] + mu_basic_terms[1]]
                    if order > 1:
                        mu_terms.append(mu_basic_terms[2])
                    if order > 2:
                        mu_terms.append(mu_basic_terms[3])
                else:
                    raise ValueError("don't know how to interpret dipole partitioning {}".format(partitioning))

                # define out reps based on partitioning style
                with logger.block(tag="axis: {}".format(a)):
                    logger.log_print(
                        [
                            "non-zero dipole terms: {nt}"
                        ],
                        nt=len(mu_terms) - mu_terms.count(0)
                    )

                    with logger.block(tag="calculating corrections..."):
                        start = time.time()
                        for q in range(order):  # total quanta
                            # logger.log_print("calculating corrections at order {q}...", q=q)
                            terms_depert = []
                            terms_deg = []
                            # should do this smarter
                            for i, j, k in ip.product(range(q + 1), range(q + 1), range(q + 1)):
                                if i + j + k == q:
                                    if len(mu_terms) <= k or len(corr_terms) <= i or len(corr_terms) <= j:
                                        new = np.zeros((len(low_spec), len(up_spec)))
                                        new_deg = new
                                    else:
                                        m = mu_terms[k]
                                        if self._is_zero(m):
                                            # to make it easy to zero stuff out
                                            new = np.zeros((len(lower_states_input), len(upper_states_input)))
                                            new_deg = new
                                        else:
                                            c_lower = corr_terms_lower[i]
                                            c_upper = corr_terms_upper[j]
                                            num = c_lower.dot(m)
                                            new = num.dot(c_upper)
                                            if degenerate_transformation is not None:
                                                new_deg = new
                                                if deg_transf_lower is not None:
                                                    new_deg = deg_transf_lower.dot(new_deg)
                                                if deg_transf_upper is not None:
                                                    new_deg = new_deg.dot(deg_transf_upper)
                                            else:
                                                new_deg = None
                                        if isinstance(new, SparseArray):
                                            new = new.asarray()
                                        new = new.reshape((len(lower_states_input), len(upper_states_input)))
                                    terms_depert.append(new)

                                    if degenerate_transformation is not None:
                                        if isinstance(new_deg, SparseArray):
                                            new_deg = new_deg.asarray()
                                        new_deg = new_deg.reshape((len(lower_states_input), len(upper_states_input)))
                                        terms_deg.append(new_deg)
                                    # raise Exception(new.toarray())
                            # print(q, a)
                            transition_moment_components[q][a] = terms_depert
                            if degenerate_transformation is not None:
                                transition_moment_components_deg[q][a] = terms_deg
                            end = time.time()
                            logger.log_print(
                                "took {t:.3f}s",
                                t=end - start
                            )

            # we calculate it explicitly like this up front in case we want to use it later since the shape
            # can be a bit confusing ([0-order, 1-ord, 2-ord], [x, y, z])
            tmom = self._compute_tmom_to_order(transition_moment_components, order-1)

            if degenerate_transformation is not None:
                tmom_deg = self._compute_tmom_to_order(transition_moment_components_deg, order-1)
            else:
                tmom_deg = None

            # mu_reps already organized like x/y/z
            # mu_reps = [
            #     [m[a] for m in mu_reps]
            #     for a in range(3)
            # ]

        # print(
        #     (tmom, transition_moment_components),
        #     (tmom_deg, transition_moment_components_deg)
        #
        # )
        with self.results:
            if degenerate_transformation is not None:
                self.results["transition_moments"] = transition_moment_components_deg
                self.results["nondegenerate_transition_moments"] = transition_moment_components
            else:
                self.results["transition_moments"] = transition_moment_components

        return [(tmom, transition_moment_components), (tmom_deg, transition_moment_components_deg), mu_reps, (corr_terms_lower, corr_terms_upper)]

    @classmethod
    def _compute_tmom_to_order(cls, transition_moment_components, order):
        # we calculate it explicitly like this up front in case we want to use it later since the shape
        # can be a bit confusing ([0-order, 1-ord, 2-ord], [x, y, z])
        return [
            sum(
                sum(ip.chain(transition_moment_components[i][j]))
                if not isinstance(transition_moment_components[i][j], (float, int, np.floating, np.integer))
                else transition_moment_components[i][j]
                for i in range(order + 1)  # correction order
            ) for j in range(3)  # xyz
        ]

    def _operator_representation(self,
                                 ders,
                                 M,
                                 space,
                                 bra_spaces,
                                 ket_spaces,
                                 order,
                                 rep_inds,
                                 base_sym="O"
                                 ):

        def get_states(m, k):
            return self._get_rep_states(m, k, order, ket_spaces, bra_spaces)

        if (
            isinstance(ders, (np.ndarray, SparseArray)) and ders.shape == (M, M)
            or self._is_zero(ders)
        ): # we were passed representations to reuse
            terms = ders
        else:

            coeffs = []
            n = self.corrs.states.ndim
            for i in range(order):
                if self._is_zero(ders):
                    coeffs.append(0)#np.zeros((n,) * i))
                else:
                    if len(ders) <= i or self._is_zero(ders[i]):
                        coeffs.append(0)#np.zeros((n,) * i))
                    else:
                        coeffs.append(ders[i])

            reps = []
            for i, mu in enumerate(coeffs):
                if self._is_zero(mu):
                    reps.append(mu)
                else:
                    m = self.get_Mi(i, mu, base_sym=base_sym)
                    inds = get_states(m, i)

                    if i >= len(rep_inds):
                        for d in range(i-len(rep_inds) + 1):
                            rep_inds.append(None)
                    rep_inds[i] = inds
                    reps.append(m)

            terms = [None] * len(reps)
            tb = self.corrs.total_basis
            for i, h in enumerate(reps):
                if self._is_zero(h):
                    terms[i] = 0
                    continue

                m_pairs = rep_inds[i]
                if m_pairs is None:
                    raise ValueError("representation indices haven't filled enough to calculate {}".format(h))

                with self.logger.block(tag="building {}".format(h)):
                    start = time.time()
                    sub = h.get_representation_matrix(m_pairs, tb
                                                      , zero_element_warning=False # expect zeros in M(3)?
                                                      # , remove_duplicates=False
                                                      )
                    end = time.time()
                    self.logger.log_print('took {t:.3f}s...', t=end - start)
                # sub = generate_rep(h, m_pairs)
                terms[i] = sub

        return terms

    # @profile
    def _operator_corrections(self,
        operator,
        correction_terms=None,
        lower_states=None,
        excited_states=None,
        degenerate_transformation=None,
        order=None
        ):
        """
        Calculates operator corrections between the wavefunctions stored.

        :param operator: operator components (1st, 2nd, 3rd derivatives in normal modes)
        :type operator: Iterable[np.ndarray]
        :return:
        :rtype:
        """

        if order is None:
            exp_order = None if 'expansion_order' not in self.expansion_options else self.expansion_options['expansion_order']
            if exp_order is None:
                exp_order = self.order
            elif isinstance(exp_order, int):
                exp_order = exp_order + 1
            else:
                if 'default' in exp_order:
                    exp_order = exp_order['default'] + 1
                else:
                    exp_order = self.order
            order = min([exp_order, self.order])

        logger = self.logger
        with logger.block(tag="Calculating operator corrections:", printoptions={'linewidth': int(1e8)}):

            space = self.corrs.coupled_states
            # store this for handling the degenerate_transformation
            lower_states_input = self.initial_state_indices if lower_states is None else lower_states
            upper_states_input = tuple(range(space.nstates)) if excited_states is None else excited_states

            if lower_states is None:
                low_spec = self.initial_state_indices
            else:
                if isinstance(lower_states, (int, np.integer)):
                    lower_states = (lower_states,)
                low_spec = lower_states
            lower_states = space[(low_spec,)]

            if excited_states is None:
                up_spec = tuple(range(space.nstates))
                excited_states = space
            else:
                if isinstance(excited_states, (int, np.integer)):
                    excited_states = (excited_states,)
                up_spec = excited_states
                excited_states = space[(up_spec,)]

            bra_spaces = lower_states  # if isinstance(lower_states, BasisStateSpace) else lower_states.to_single().take_unique()
            ket_spaces = excited_states  # if isinstance(excited_states, BasisStateSpace) else excited_states.to_single().take_unique()

            # M = len(space.indices)
            logger.log_print(
                [
                    "lower/upper states: {l}/{u}"
                ],
                l=len(low_spec),
                u=len(up_spec)
            )

            if correction_terms is None:

                with logger.block(tag='taking correction subspaces:'):
                    start = time.time()
                    logger.log_print('getting ground space...')
                    corr_terms_lower = [self.corrs.wfn_corrections[i][low_spec, :] for i in range(order)]
                    logger.log_print(
                        [
                            "highest-order space: {s}",
                        ],
                        s=corr_terms_lower[-1]
                    )

                    logger.log_print('getting excited space...')
                    corr_terms_upper = [self.corrs.wfn_corrections[i][up_spec, :].T for i in range(order)]
                    logger.log_print(
                        [
                            "highest-order space: {s}",
                        ],
                        s=corr_terms_upper[-1]
                    )

                    end = time.time()
                    logger.log_print(
                        [
                            "took {t:.3f}s..."
                        ],

                        t=end - start
                    )
            else:
                corr_terms_lower, corr_terms_upper = correction_terms

            rep_inds = [None] * 4
            M = self.corrs.wfn_corrections[0].shape[1]
            total_space = self.corrs.total_basis
            components = np.zeros((order,)).tolist()  # x, y, and z components of the 0th, 1st, and 2nd order stuff

            # raise Exception(corr_terms[1], corr_terms_lower[1], corr_terms_upper[1], low_spec, len(low_spec), len(up_spec))

            terms = self._operator_representation(
                operator,
                M,
                total_space,
                bra_spaces,
                ket_spaces,
                order,
                rep_inds,
                base_sym="O"
            )

            # raise Exception("...")
            # logger.log_print(
            #     [
            #         "took {t:.3f}s..."
            #     ],
            #     t=end-start
            # )

            # raise Exception(mu_terms)
            if degenerate_transformation is None:
                degenerate_transformation = self.corrs.degenerate_transf
            if degenerate_transformation is not None:
                # raise  Exception("ugh")
                # degenerate_transformation = degenerate_transformation.T

                if len(lower_states_input) > 1:
                    deg_transf_lower = degenerate_transformation[np.ix_(lower_states_input, lower_states_input)]
                    if self.degenerate_transformation_layout == "column":
                        deg_transf_lower = deg_transf_lower.T
                else:
                    deg_transf_lower = None
                if len(upper_states_input) > 1:
                    if self.degenerate_transformation_layout == "column":
                        deg_transf_upper = degenerate_transformation[:, upper_states_input]
                    else:
                        deg_transf_upper = degenerate_transformation[upper_states_input, :].T
                else:
                    deg_transf_upper = None
            else:
                deg_transf_lower = deg_transf_upper = None

            if degenerate_transformation is not None:
                components_deg = np.zeros((order,)).tolist()  # x, y, and z components of the 0th, 1st, and 2nd order stuff
            else:
                components_deg = None

            corr_terms = self.corrs.wfn_corrections
            # define out reps based on partitioning style
            logger.log_print(
                [
                    "non-zero terms: {nt}"
                ],
                nt=len(terms) - terms.count(0)
            )

            with logger.block(tag="calculating corrections..."):
                start = time.time()
                for q in range(order):  # total quanta
                    # logger.log_print("calculating corrections at order {q}...", q=q)
                    terms_depert = []
                    terms_deg = []
                    # should do this smarter
                    for i, j, k in ip.product(range(q + 1), range(q + 1), range(q + 1)):
                        if i + j + k == q:
                            if len(terms) <= k or len(corr_terms) <= i or len(corr_terms) <= j:
                                new = np.zeros((len(low_spec), len(up_spec)))
                                new_deg = new
                            else:
                                m = terms[k]
                                if isinstance(m, (int, float, np.integer, np.floating)) and m == 0:
                                    # to make it easy to zero stuff out
                                    new = np.zeros((len(lower_states_input), len(upper_states_input)))
                                    new_deg = new
                                else:
                                    c_lower = corr_terms_lower[i]
                                    c_upper = corr_terms_upper[j]
                                    num = c_lower.dot(m)
                                    new = num.dot(c_upper)
                                    if degenerate_transformation is not None:
                                        new_deg = new
                                        if deg_transf_lower is not None:
                                            new_deg = deg_transf_lower.dot(new_deg)
                                        if deg_transf_upper is not None:
                                            new_deg = new_deg.dot(deg_transf_upper)
                                    else:
                                        new_deg = None
                                if isinstance(new, SparseArray):
                                    new = new.asarray()
                                new = new.reshape((len(lower_states_input), len(upper_states_input)))
                            terms_depert.append(new)

                            if degenerate_transformation is not None:
                                if isinstance(new_deg, SparseArray):
                                    new_deg = new_deg.asarray()
                                new_deg = new_deg.reshape(
                                    (len(lower_states_input), len(upper_states_input)))
                                terms_deg.append(new_deg)
                            # raise Exception(new.toarray())
                    # print(q, a)
                    components[q] = terms_depert
                    if degenerate_transformation is not None:
                        components[q] = terms_deg
                    end = time.time()
                    logger.log_print(
                        "took {t:.3f}s",
                        t=end - start
                    )

            # we calculate it explicitly like this up front in case we want to use it later since the shape
            # can be a bit confusing ([0-order, 1-ord, 2-ord], [x, y, z])
            val = self._compute_corr_to_order(components, order - 1)

            if degenerate_transformation is not None:
                val_deg = self._compute_corr_to_order(components_deg, order - 1)
            else:
                val_deg = None

        return [
            (val, components),
            (val_deg, components_deg),
            terms,
            (corr_terms_lower, corr_terms_upper)
        ]

    @classmethod
    def _compute_corr_to_order(cls, components, order):
        # we calculate it explicitly like this up front in case we want to use it later since the shape
        # can be a bit confusing ([0-order, 1-ord, 2-ord], [x, y, z])
        return sum(
                sum(ip.chain(components[i]))
                    if not isinstance(components[i], (float, int, np.floating, np.integer)) else
                components[i]
                for i in range(order+1)  # correction order
            )



    class TermHolder(tuple):
        """symbolic wrapper on Tuple so we can know that we've canonicalized some term"""
    @property
    def dipole_terms(self):
        if self._dipole_terms is None:
            if 'dipole_terms' in self.expansion_options:
                self._dipole_terms = self.TermHolder(self.expansion_options['dipole_terms'])
            else:
                self._dipole_terms = self.TermHolder(
                    DipoleTerms(self.mol, modes=self.modes, mode_selection=self.mode_selection,
                                **ParameterManager(self.expansion_options).filter(DipoleTerms)
                                ).get_terms()
                )
        elif not isinstance(self._dipole_terms, self.TermHolder):
            self._dipole_terms = self.TermHolder(
                DipoleTerms(self.mol, modes=self.modes, mode_selection=self.mode_selection,
                            dipole_derivatives=self._dipole_terms,
                            **ParameterManager(self.expansion_options).filter(DipoleTerms)
                            ).get_terms()
            )
        # print([x for x in self._dipole_terms])
        return self._dipole_terms

    @dipole_terms.setter
    def dipole_terms(self, v):
        self._dipole_terms = v
        self._tm_dat = None

    class DipolePartitioningMethod(enum.Enum):
        Standard="standard"
        Intuitive="intuitive"

    @property
    def dipole_partitioning(self):
        return self._dipole_partitioning

    @dipole_partitioning.setter
    def dipole_partitioning(self, p):
        part = self.DipolePartitioningMethod(p)
        if part is not self._dipole_partitioning:
            self._dipole_partitioning = part
            if self._tm_dat is not None:
                self._tm_dat[0] = None

    def _load_tm_dat(self):
        if self._tm_dat is None:
            self._tm_dat = self._transition_moments(*self.dipole_terms)
        elif self._tm_dat[0] is None:
            self._tm_dat = self._transition_moments(*self._tm_dat[2], correction_terms=self._tm_dat[3])

    @property
    def transition_moments(self):
        """
        Computes the transition moments between wavefunctions stored in the object

        :return:
        :rtype:
        """

        self._load_tm_dat()
        if self._tm_dat[1][0] is not None:
            return self._tm_dat[1][0]
        else:
            return self._tm_dat[0][0]

    @property
    def deperturbed_transition_moments(self):
        """
        Computes the transition moments between wavefunctions stored in the object

        :return:
        :rtype:
        """

        self._load_tm_dat()
        return self._tm_dat[0][0]

    @property
    def transition_moment_corrections(self):
        """
        Computes the transition moment corrections between wavefunctions stored in the object

        :return:
        :rtype:
        """

        self._load_tm_dat()
        if self._tm_dat[1][0] is not None:
            return self._tm_dat[1][1]
        else:
            return self._tm_dat[0][1]

    @property
    def deperturbed_transition_moment_corrections(self):
        """
        Computes the transition moment corrections between wavefunctions stored in the object

        :return:
        :rtype:
        """

        self._load_tm_dat()
        return self._tm_dat[0][1]

    @property
    def oscillator_strengths(self):
        """
        Computes the oscillator strengths for transitions from the ground state to the other states

        :return:
        :rtype:
        """

        tms = self.transition_moments
        return self._oscillator_strengths(tms)

    @property
    def deperturbed_oscillator_strengths(self):
        """
        Computes the oscillator strengths for transitions from the ground state to the other states

        :return:
        :rtype:
        """

        tms = self.deperturbed_transition_moments
        return self._oscillator_strengths(tms)

    def oscillator_strengths_to_order(self, order):
        """

        :param tms:
        :type tms:
        :return:
        :rtype:
        """

        return self._oscillator_strengths(self._compute_tmom_to_order(
            self.transition_moment_corrections,
            order
        ))

    def deperturbed_oscillator_strengths_to_order(self, order):
        """

        :param tms:
        :type tms:
        :return:
        :rtype:
        """

        return self._oscillator_strengths(self._compute_tmom_to_order(
            self.deperturbed_transition_moment_corrections,
            order
        ))

    def _oscillator_strengths(self, tms):

        res = {}
        for n,idx in enumerate(self.initial_state_indices):
            gs_tms = np.array([tms[i][n] for i in range(3)]).T
            osc = np.linalg.norm(gs_tms, axis=1) ** 2
            res[idx] = osc
        if self.initial_states is None:
            res = res[0]
        return res

    @property
    def intensities(self):
        """
        Computes the intensities for transitions from the ground state to the other states

        :return:
        :rtype:
        """

        freqs, ints = self._intensities(self.oscillator_strengths)
        with self.results: # TODO: this really isn't the right place for this...
            self.results['spectrum'] = {"state_{}".format(i):(f,ints[i]) for i,f in freqs.items()}

        return ints

    @property
    def deperturbed_intensities(self):
        """
        Computes the intensities for transitions from the ground state to the other states

        :return:
        :rtype:
        """
        return self._dep_intensities(self.deperturbed_oscillator_strengths)[1]

    def intensities_to_order(self, order, return_freqs=False):
        """
        Computes the intensities for transitions from the ground state to the other states

        :return:
        :rtype:
        """
        res = self._intensities(self.oscillator_strengths_to_order(order), energy_order=order)
        if not return_freqs:
            res = res[1]
        return res

    def deperturbed_intensities_to_order(self, order, return_freqs=False):
        """
        Computes the intensities for transitions from the ground state to the other states

        :return:
        :rtype:
        """
        res = self._dep_intensities(self.deperturbed_oscillator_strengths_to_order(order), energy_order=order)
        if not return_freqs:
            res = res[1]
        return res

    def _get_int(self, eng, oscs):
        units = UnitsData.convert("OscillatorStrength", "KilometersPerMole")
        if isinstance(oscs, dict):
            freqs = {}
            ints = {}
            for idx, o in oscs.items():
                freqs[idx] = eng - eng[idx]
                ints[idx] = units * freqs[idx] * o
        else:
            freqs = eng - eng[0]
            ints = units * freqs * oscs
        return freqs, ints
    def _dep_intensities(self, oscs, energy_order=None):
        if energy_order is None:
            eng = self.deperturbed_energies
        else:
            eng = self.energies_to_order(energy_order)

        freqs, ints = self._get_int(eng, oscs)
        if self.initial_states is None:
            freqs = freqs[0]
            ints = ints[0]
        return freqs, ints

    def _intensities(self, oscs, energy_order=None):
        if energy_order is None:
            eng = self.energies
        elif self.degenerate_transformation is not None:
            raise NotImplementedError("degenerate transformation not yet included for energies_to_order...")
        else:
            eng = self.energies_to_order(energy_order)

        freqs, ints = self._get_int(eng, oscs)
        if self.initial_states is None:
            freqs = freqs[0]
            ints = ints[0]
        return freqs, ints

    @property
    def zero_order_intensities(self):
        """
        Computes the harmonic intensities for transitions from the ground state to the other states

        :return:
        :rtype:
        """
        freqs, ints = self.deperturbed_intensities_to_order(0, return_freqs=True)

        with self.results: #TODO: this is the wrong place for this
            self.results['zero_order_spectrum'] = {"state_{}".format(i):(f,ints[i]) for i,f in freqs.items()}

        return ints

    def prep_operator_terms(self, coeffs):
        needs_transform = None
        if isinstance(coeffs, dict):
            needs_transform = coeffs.get('transformed', None)
            coeffs = coeffs['expansion']
        coeffs = [
            x
                if isinstance(x, (int, np.integer, float, np.floating)) else
            float(x)
                if (isinstance(x, np.ndarray) and x.shape == ()) else
            np.asanyarray(x)
            for x in coeffs
        ]
        if needs_transform is None:
            needs_transform = any(
                isinstance(x, np.ndarray) and x.shape[-1] > self.corrs.states.ndim
                for x in coeffs
            )
        if needs_transform:
            const = coeffs[0]
            coeffs = OperatorTerms(self.mol, modes=self.modes, mode_selection=self.mode_selection,
                                   operator_derivatives=coeffs[1:],
                                   **ParameterManager(self.expansion_options).filter(OperatorTerms)
                                   ).get_terms()
            coeffs = [const] + coeffs
        return coeffs

    def operator_correction_data(self, operator_coeffs, order=None):
        if isinstance(operator_coeffs, dict):
            res = {}
            corr_terms = None
            for key, coeffs in operator_coeffs.items():
                coeffs = self.prep_operator_terms(coeffs)
                operator_dat = self._operator_corrections(
                    coeffs,
                    correction_terms=corr_terms,
                    order=order
                )
                corr_terms = operator_dat[3]
                res[key] = operator_dat
            return res
        else:
            return self._operator_corrections(
                operator_coeffs,
                order=order
            )

    def generate_intensity_breakdown(self, include_wavefunctions=True):
        """
        Generates a breakdown of the terms that contribute to the intensity
        Returns in a format that can be directly exported to JSON if desired.

        :return:
        :rtype:
        """
        raise NotImplementedError("unmaintained")

        from collections import OrderedDict

        engs = self.energies
        freqs = engs - engs[0]
        terms = OrderedDict((
            ("frequencies", freqs.tolist()),
            ('axes', self.mol.inertial_axes.tolist()),
            ("states", self.corrs.states.excitations),
            ('breakdowns', OrderedDict())
        ))

        if include_wavefunctions:
            terms["wavefunctions"] = {
                "corrections": [x.toarray().tolist() for x in self.corrs.wfn_corrections],
                "coupled_states": self.corrs.total_basis.excitations
            }
        bds = terms['breakdowns']

        # dts = self.dipole_terms

        dipole_breakdowns = OrderedDict((
            ("Full",      (True, True, True, True)),
            ("Constant",  (True, False, False, False)),
            ("Linear",    (False, True, False, False)),
            ("Quadratic", (False, False, True, False)),
            ("Cubic",     (False, False, False, True)),
            ("Order0",    (True, True, False, False)),
            ("Order1",    (True, True, True, False))
            # ("Order2",  (True, True, True, True)),
        ))

        # wfn_terms.append(freqs.tolist())
        for key in dipole_breakdowns:
            if self.logger is not None:
                self.logger.log_print(
                    "-" * 50 + "{k} {m}" + "-" * 50,
                    k=key,
                    m=self.dipole_partitioning.value,
                    padding="",
                    newline=" "
                )
            if key == "Full":
                # gotta come first
                ints = self.intensities
            else:
                mu_reps = self._tm_dat[2]
                dts = dipole_breakdowns[key]
                new_mu = [
                    [ m[i] if d and len(m) > i else 0 for i, d in enumerate(dts)]
                    for m in mu_reps # x/y/z
                ]
                tms = self._transition_moments(*new_mu)
                oscs = self._oscillator_strengths(tms[0])
                freqs, ints = self._intensities(oscs)

            full_corrs = self.transition_moment_corrections
            corrs_keys = [y for x in ip.chain(
                [(i, j, k) for i, j, k in ip.product(range(q + 1), range(q + 1), range(q + 1)) if i + j + k == q]
                for q in range(len(full_corrs))
            ) for y in x]

            full_corrs = np.concatenate([np.array(x) for x in full_corrs], axis=1)
            # transpose so we have x-y-z as outermost dimension and correction as innermost
            full_corrs = full_corrs.transpose((0, 2, 3, 1)).tolist()

            bds[key] = OrderedDict((
                ("intensities", ints),
                ("corrections", {
                    "keys": corrs_keys,
                    "values": OrderedDict(zip(['X', 'Y', 'Z'], full_corrs))
                })
            ))

        return terms

    @classmethod
    def write_CSV_breakdown(cls, file, intensity_breakdown, padding=None):
        """
        Writes an intensity breakdown to a CSV by annoyingly flattening all the arrays

        :param file:
        :type file:
        :param intensity_breakdown:
        :type intensity_breakdown:
        :return:
        :rtype:
        """
        raise NotImplementedError("unmaintained")

        import csv

        h2w = UnitsData.convert("Hartrees", "Wavenumbers")

        if padding is None:
            padding = []
        else:
            padding = list(padding)

        if isinstance(file, str):
            with open(file, "w+") as fp:
                cls.write_CSV_breakdown(fp, intensity_breakdown)
        else:
            writer = csv.writer(file)
            # write header
            ib = intensity_breakdown
            bds = ib['breakdowns']
            keys = ['State', 'Freq.'] + list(bds.keys())
            writer.writerows([padding + keys])

            # write main intensity table
            states = ib['states']
            freqs = h2w * np.array(ib['frequencies'])
            intensity_table = []
            for i, s in enumerate(states):
                f = freqs[i]
                ints = [b['intensities'][i] for b in bds.values()]
                intensity_table.append([s, f] + ints)
            writer.writerows([padding + x for x in intensity_table])

            for k, b in bds.items():
                corr_block = [[k]]
                corrs = b['corrections']
                corr_keys = ["<gs^({0})|mu^({2})|es^({1})>".format(*lab) for lab in corrs['keys']]
                for ax, vals in corrs['values'].items():
                    corr_block.append([ax])
                    corr_block.append([" "] + corr_keys)
                    for s, v in zip(states, vals[0]):
                        corr_block.append([s] + list(v))
                writer.writerows([padding + x for x in corr_block])

            if 'wavefunctions' in ib:

                coupled_state_blocks = [["Wavefunction Corrections"]]
                wfn_corrs = np.array(ib['wavefunctions']["corrections"])
                coupled_states = ib['wavefunctions']["coupled_states"]

                cs_keys = list(coupled_states)
                for corr_block, state in zip(wfn_corrs.transpose((1, 0, 2)), states):
                    coupled_state_blocks.append([state] + cs_keys)
                    for i, b in enumerate(corr_block):
                        coupled_state_blocks.append(["|n^({})>".format(i)] + list(b))

                writer.writerows([padding + x for x in coupled_state_blocks])

    @classmethod
    def _format_energies_table(cls, gs, zpe, states, freqs, real_fmt='{:>12.5f}'):

        n_modes = len(gs)

        zpe = np.asanyarray(zpe)
        freqs = np.asanyarray(freqs)

        ncols = freqs.shape[1] if len(freqs.shape) > 0 else 1
        states_fmt = '{:<1.0f} ' * n_modes
        padding = len(states_fmt.format(*[0 for _ in range(n_modes)]))
        num_digits = len(real_fmt.format(0))
        dash_fmt = '{:>' + str(num_digits) + "}"
        htag = "Harmonic"; hpad = (num_digits - len(htag))
        atag = "Anharmonic"; apad = (num_digits - len(atag))
        bar = (
                (" " * hpad if hpad > 0 else "") + htag
                + " "
                + (" " * apad if apad > 0 else "") + atag
                )
        if zpe is not None:
            header = ["State" + " " * (padding - len("State")) + bar + " " + bar]
            header += [
                " " * padding
                + " " * (num_digits - 3) + "ZPE"
                + " " + " " * (num_digits - 3) + "ZPE"
                + " "
                + " " * (num_digits - 9) + "Frequency"
                + " " + " " * (num_digits - 9) + "Frequency"
            ]
            zpe_fmt = (real_fmt + " ") * ncols + (dash_fmt + " ") * ncols
            freq_fmt = (dash_fmt + " ") * ncols + (real_fmt + " ") * ncols
            lines = [
                        states_fmt.format(*gs) + zpe_fmt.format(*zpe, *["-"] * ncols)] + [
                        states_fmt.format(*s) + freq_fmt.format(*["-"] * ncols, *e)
                            for s, e in
                        zip(states, freqs)
                    ]
        else:
            header = ["State" + " " * (padding - len("State")) + bar]
            header += [
                " " * padding
                + " " * (num_digits - 9) + "Frequency"
                + " " + " " * (num_digits - 9) + "Frequency"
            ]
            freq_fmt = ncols + (real_fmt + " ") * ncols
            lines = [
                states_fmt.format(*s) + freq_fmt.format(*e)
                    for s, e in
                zip(states, freqs)
            ]

        return "\n".join([
            *header,
            *lines,
        ])

    def _format_group_energies_table(self, engs, states=None, zpe=None, freqs=None, real_fmt='{:>12.5f}'):

        if states is None:
            states = self.corrs.states.excitations

        h2w = UnitsData.convert("Hartrees", "Wavenumbers")

        res = {}

        harm_engs = self.zero_order_energies * h2w
        engs = engs * h2w

        if zpe is not None:
            if self.initial_states is not None:
                raise NotImplementedError("don't know with multiple initial states and a passed ZPE")
            return self._format_energies_table(
                [0]*states.shape[-1],
                zpe,
                states[1:],
                freqs,
                real_fmt=real_fmt
            )

        for init in self.initial_state_indices:
            subsel = np.concatenate([np.arange(init), np.arange(init+1, len(states))])

            # if zpe is None:# and np.all(states[0] == 0):
            zpe = [harm_engs[init], engs[init]]

            # if freqs is None:
            harm_freq = harm_engs[subsel] - harm_engs[init]
            anh_freqs = engs[subsel] - engs[init]
            freqs = np.column_stack([harm_freq, anh_freqs])

            res[init] = self._format_energies_table(
                states[init],
                zpe,
                states[subsel],
                freqs,
                real_fmt=real_fmt
            )

        if self.initial_states is None:
            res = res[0]
        else:
            max_line_len = max(len(l) for l in next(iter(res.values())).splitlines())
            sep = "\n"+("-"*max_line_len)+"\n"
            res = sep.join(res.values())

        return res

    def format_energies_table(self, states=None, zpe=None, freqs=None, real_fmt='{:>12.5f}'):
        return self._format_group_energies_table(
            self.energies,
            states=states,
            zpe=zpe,
            freqs=freqs,
            real_fmt=real_fmt
        )

    def format_deperturbed_energies_table(self, states=None, zpe=None, freqs=None, real_fmt='{:>12.5f}'):
        return self._format_group_energies_table(
            self.deperturbed_energies,
            states=states,
            zpe=zpe,
            freqs=freqs,
            real_fmt=real_fmt
        )

    def format_property_matrices(self, states, prop_corrs, real_fmt="{:>.8e}", padding_fmt='{:>16}'):

        # reshape data so it's one big matrix...
        order = len(prop_corrs)
        res = {}
        for n,idx in enumerate(self.initial_state_indices):

            term_keys = []
            terms = []

            for q in range(order):  # total quanta
                for pq in prop_corrs[q]:
                    # if len(pq) > 1:
                    #     raise NotImplementedError(
                    #         "don't know what to do when more than one starting state (got {})".format(
                    #             len(pq)
                    #         )
                    #     )
                    terms.append(pq[n])
                for i, j, k in ip.product(range(q + 1), range(q + 1), range(q + 1)):
                    if i + j + k == q:
                        term_keys.append("<{i}|M({k})|{j}>".format(i=i, j=j, k=k))

            # print(len(terms), [len(x) for x in terms])
            terms = np.array(terms).T

            nels = len(padding_fmt.format(real_fmt.format(0)))
            header_fmt = "{:<" + str(nels) + "}"
            header = " ".join(header_fmt.format(k) for k in term_keys)

            n_modes = self.corrs.total_basis.ndim
            padding = np.max([len(str("0 " * n_modes)), 1]) + 1
            header = "State" + " " * (padding - 3) + header

            n_terms = len(term_keys)
            prop_mat = header + "\n" + "\n".join(
                "  " + ("{:<1.0f} " * n_modes).format(*s) +
                " ".join(padding_fmt.format(real_fmt.format(ll)) for ll in l)
                for s, l in zip(states, terms)
            )

            res[idx] = prop_mat

        if self.initial_states is None:
            res = res[0]
        else:
            _ = []
            for idx, mats in res.items():
                head = "Initial State: {}\n".format("{:<1.0f} " * len(states[idx])).format(*states[idx])
                _.append(head + mats)
            max_line_len = max(len(l) for l in _[0].splitlines())
            sep = "\n"+("-"*max_line_len)+"\n"
            res = sep.join(_)

        return res

    def format_dipole_contribs_tables(self):

        states = self.corrs.states.excitations
        corrs = self.transition_moment_corrections
        mats = []
        for a in range(3):
            prop = [x[a] for x in corrs]
            mats.append(self.format_property_matrices(states, prop))

        return mats

    def format_deperturbed_dipole_contribs_tables(self):

        states = self.corrs.states.excitations
        corrs = self.deperturbed_transition_moment_corrections
        mats = []
        for a in range(3):
            prop = [x[a] for x in corrs]
            mats.append(self.format_property_matrices(states, prop))

        return mats

    def format_energy_corrections_table(self, real_fmt="{:>12.5f}"):

        h2w = UnitsData.convert("Hartrees", "Wavenumbers")
        states = self.corrs.states.excitations
        e_corrs = self.corrs.all_energy_corrs #ordered like (state, order, corr_list)
        order = len(e_corrs[0])
        term_keys = []
        for k in range(1, order):
            if e_corrs[0][k] is not None: # from odd order terms...
                for i in range(0, k):
                    term_keys.append("<0|dH({j})|{i}>".format(j=k-i, i=i))
        terms = [ [y for y in x if y is not None] for x in e_corrs ]
        terms = [ h2w*np.concatenate(tl) for tl in terms if len(tl) > 0 ]

        nels = len(real_fmt.format(0))
        header_fmt = "{:<" + str(nels) + "}"
        header = " ".join(header_fmt.format(k) for k in term_keys)

        n_modes = self.corrs.total_basis.ndim
        padding = np.max([len(str("0 " * n_modes)), 1]) + 1
        header = "State" + " " * (padding - 3) + header

        prop_mat = header + "\n" + "\n".join(
            "  " + ("{:<1.0f} " * n_modes).format(*s) +
            " ".join(real_fmt.format(ll) for ll in l)
            for s, l in zip(states, terms)
        )

        return prop_mat

    def _format_intensities_table(self, states, freqs, ints, harm_freqs, harm_ints, real_fmt="{:>12.5f}"):
        # simple utility function pulled from the unit tests

        n_modes = self.corrs.total_basis.ndim
        padding = np.max([len(str("0 " * n_modes)), 1]) + 1

        num_digits = len(real_fmt.format(0))
        ftag = "Frequency"; fpad = (num_digits - len(ftag))
        itag = "Intensity"; ipad = (num_digits - len(itag))
        bar = (
            (" " * fpad if fpad > 0 else "") + ftag
            + " "
            + (" " * ipad if ipad > 0 else "") + itag
        )

        spacer = "    "
        report = (
                    " " * (padding + 2) + " " * (num_digits - 2) + "Harmonic" +
                    " " * (num_digits - len("Harmonic")) + spacer
                    + " " * (num_digits - 2) + "Anharmonic\n"
                    + "State" + " " * (padding - 3) + bar + spacer + bar + "\n"
                 ) + "\n".join(
            (
                (
                        "  " + ("{:<1.0f} " * n_modes) + " "
                        + real_fmt + " " + real_fmt
                        + spacer
                        + real_fmt + " " + real_fmt
                ).format(*s, hf, hi, f, i)
                for s, hf, hi, f, i in zip(states, harm_freqs, harm_ints, freqs, ints)
            ))

        return report

    def _format_group_intensities_table(self, engs, ints, states=None, real_fmt='{:>12.5f}'):

        if states is None:
            states = self.corrs.states.excitations

        h2w = UnitsData.convert("Hartrees", "Wavenumbers")

        res = {}

        harm_engs = self.zero_order_energies * h2w
        engs = engs * h2w

        zints = self.zero_order_intensities

        for init_idx in self.initial_state_indices:
            subsel = np.concatenate([np.arange(init_idx), np.arange(init_idx+1, len(states))])

            harm_freqs = harm_engs[subsel] - harm_engs[init_idx]
            freqs = engs[subsel] - engs[init_idx]

            if isinstance(zints, dict):
                harm_ints = zints[init_idx][subsel]
            else:
                harm_ints = zints

            if isinstance(ints, dict):
                anh_ints = ints[init_idx][subsel]
            else:
                anh_ints = ints

            res[init_idx] = self._format_intensities_table(states[subsel], freqs, anh_ints, harm_freqs, harm_ints, real_fmt=real_fmt)
            # self._format_energies_table(
            #     states[init],
            #     zpe,
            #     states[subsel],
            #     freqs,
            #     real_fmt=real_fmt
            # )

        if self.initial_states is None:
            res = res[0]
        else:
            _ = []
            for idx, mats in res.items():
                head = "Initial State: {}\n".format("{:<1.0f} " * len(states[idx])).format(*states[idx])
                _.append(head + mats)
            max_line_len = max(len(l) for l in _[0].splitlines())
            sep = "\n"+("-"*max_line_len)+"\n"
            res = sep.join(_)

        return res

    def format_intensities_table(self, real_fmt="{:>12.5f}"):
        return self._format_group_intensities_table(
            self.energies,
            self.intensities,
            states=self.corrs.states.excitations,
            real_fmt=real_fmt
        )

    def format_deperturbed_intensities_table(self, real_fmt="{:>12.5f}"):
        return self._format_group_intensities_table(
            self.deperturbed_energies,
            self.deperturbed_intensities,
            states=self.corrs.states.excitations,
            real_fmt=real_fmt
        )


    def _get_spectrum(self,
                     engs,
                      intensities
                     ):
        from ..Spectra import DiscreteSpectrum, MultiSpectrum

        # if states is None:
        states = self.corrs.states.excitations


        h2w = UnitsData.convert("Hartrees", "Wavenumbers")

        res = {}

        harm_engs = self.zero_order_energies * h2w
        engs = engs * h2w

        for init_idx in self.initial_state_indices:
            subsel = np.concatenate([np.arange(init_idx), np.arange(init_idx + 1, len(states))])

            freqs = engs[subsel] - engs[init_idx]

            if isinstance(intensities, dict):
                anh_ints = intensities[init_idx][subsel]
            else:
                anh_ints = intensities

            res[init_idx] = DiscreteSpectrum(freqs, anh_ints)

        return MultiSpectrum(list(res.values()))

    def get_spectrum(self):
        return self._get_spectrum(
            self.energies,
            self.intensities
        )
    def get_deperturbed_spectrum(self):
        return self._get_spectrum(
            self.deperturbed_energies,
            self.deperturbed_intensities
        )

    def _format_operator_table(self, states, val, zo_val, real_fmt="{:>12.5f}"):
        # simple utility function pulled from the unit tests

        n_modes = self.corrs.total_basis.ndim
        padding = np.max([len(str("0 " * n_modes)), 1]) + 1

        num_digits = len(real_fmt.format(0))

        spacer = " "
        line_template = (
                "  " + ("{:<1.0f} " * n_modes) + " "
                + real_fmt
                + spacer
                + real_fmt
        )
        htag = "Harmonic"; hpad = (num_digits - len(htag))
        atag = "Anharmonic"; apad = (num_digits - len(atag))
        bar = (
                (" " * hpad if hpad > 0 else "") + htag
                + spacer
                + (" " * apad if apad > 0 else "") + atag
                )

        report = ["State" + " " * (padding - len("State") + 2) + bar]
        report = report + [
                line_template.format(*s, hi, i)
                for s, hi, i in zip(states, zo_val, val)
            ]
        report = "\n".join(report)

        return report

    def _format_group_operator_table(self, operator_data, states=None, real_fmt='{:>12.5f}'):

        if states is None:
            states = self.corrs.states.excitations

        res = {}

        for init_idx in self.initial_state_indices:
            zo_val = self._compute_corr_to_order(operator_data[0][1], 0)[init_idx]

            if self.degenerate_transformation is None:
                val = operator_data[0][0]
            else:
                val = operator_data[1][0]
            val = val[init_idx]

            res[init_idx] = self._format_operator_table(states, val, zo_val, real_fmt=real_fmt)
            # self._format_energies_table(
            #     states[init],
            #     zpe,
            #     states[subsel],
            #     freqs,
            #     real_fmt=real_fmt
            # )

        if self.initial_states is None:
            res = res[0]
        else:
            _ = []
            for idx, mats in res.items():
                head = "Initial State: {}\n".format("{:<1.0f} " * len(states[idx])).format(*states[idx])
                _.append(head + mats)
            max_line_len = max(len(l) for l in _[0].splitlines())
            sep = "\n"+("-"*max_line_len)+"\n"
            res = sep.join(_)

        return res

    def format_operator_table(self, operators, real_fmt="{:>12.5f}"):

        if not isinstance(operators, dict):
            operators = {'Operator': operators}
        corr_data = self.operator_correction_data(operators)
        states = self.corrs.states.excitations

        res = []
        for k, dat in corr_data.items():
            res.append(
                "{sep} {k} {sep}\n{tab}".format(k=k,
                                                sep="=" * 50,
                                                tab=self._format_group_operator_table(
                                                    dat,
                                                    states=states,
                                                    real_fmt=real_fmt
                                                ))
            )
        res = "\n".join(res)

        return res



"""
Temporary storage for an attempt to get G matrix elements purely symbolically
but which lacked sufficient information about the coordinates involved
"""

import numpy as np
from .Helpers import sym, AnalyticModelBase

class OrientationVector:
    def __init__(self, x, y, z, basis):
        self.vec = [x, y, z]
        self.basis = basis
    def __neg__(self):
        return type(self)(*(-a for a in self.vec), self.basis)
    def dot(self, other:'OrientationVector'):
        if other.basis != self.basis:
            raise ValueError("bases not the same: {} vs. {}".format(self.basis, other.basis))
        else:
            return sum(x*y for x,y in zip(self.vec, other.vec))
    @classmethod
    def rotation_matrix(cls, angle, axis:'OrientationVector'):
        # a = list(axis.vec)
        # if a == [1, 0, 0] or a == [-1, 0, 0]:
        #     mat = [
        #         OrientationVector(sym.cos(angle), -sym.sin(angle), 0, axis.basis),
        #         OrientationVector(sym.sin(angle),  sym.cos(angle), 0, axis.basis),
        #         OrientationVector(             0,               0, 1, axis.basis)
        #     ]
        #         elif i == 2:
        #             mat = [
        #                 OrientationVector(1,              0,               0, axis.basis),
        #                 OrientationVector(0, sym.cos(angle), -sym.sin(angle), axis.basis),
        #                 OrientationVector(0, sym.sin(angle),  sym.cos(angle), axis.basis)
        #             ]
        #         if v == -axis:
        #             mat = [-a for a in mat]
        # elif a == [1, 0, 0] or a == [-1, 0, 0]:
        #     mat = [
        #             OrientationVector(sym.cos(angle), 0, -sym.sin(angle), axis.basis),
        #             OrientationVector(             0, 1,               0, axis.basis),
        #             OrientationVector(sym.sin(angle), 0,  sym.cos(angle), axis.basis)
        #         ]
        # else:
        #     raise Exception("...")
        a, b, c = axis.vec
        cos = sym.cos(angle)
        ncos = 1-cos
        sin = sym.sin(angle)
        # Base formula: vXv*(1-cos(a) + e3.a * sin(a)
        mat = RotationMatrix(
            OrientationVector(cos + (a**2)*ncos,  a*b*ncos - c*sin,  a*c*ncos + b*sin, axis.basis),
            OrientationVector( a*b*ncos + c*sin, cos + (b**2)*ncos,  b*c*ncos - a*sin, axis.basis),
            OrientationVector( a*c*ncos - b*sin,  b*c*ncos + a*sin, cos + (c**2)*ncos, axis.basis)
        )
        return mat
    def __repr__(self):
        return "{}({}, {})".format(type(self).__name__, self.vec, self.basis)
class RotationMatrix:
    def __init__(self, v1:OrientationVector, v2:OrientationVector, v3:OrientationVector):
        self.vectors = [v1, v2, v3]
    def dot(self, v):
        return OrientationVector(*(r.dot(v) for r in self.vectors), self.vectors[0].basis)
    def __neg__(self):
        return type(self)(*(-x for x in self.vectors))
class BondVector:
    from McUtils.Scaffolding import Logger, NullLogger
    # logger = Logger()
    logger = NullLogger()
    def __init__(self, i, j, norm=1, embedding=None):
        if i == j:
            raise ValueError("bond from {} to {} invalid".format(i, j))
        self.i = i
        self.j = j
        self.norm = norm
        self.embedding = embedding

    def one_common_atom(self, bond:'BondVector', mode='shared'):
        if self.i == bond.i or self.i == bond.j:
            if self.j == bond.i or self.j == bond.j:
                raise ValueError("all shared atoms")
            if mode == 'shared':
                return self.i
            else:
                return self.j
        elif self.j == bond.i or self.j == bond.j:
            if mode == 'shared':
                return self.j
            else:
                return self.i
        else:
            raise ValueError("no shared atoms")
    def shared_atom(self, bond:'BondVector'):
        return self.one_common_atom(bond, mode='shared')
    def unshared_atom(self, bond:'BondVector'):
        return self.one_common_atom(bond, mode='unshared')
    @property
    def direction(self):
        return type(self)(self.i, self.j, norm=1, embedding=self.embedding)
    def __repr__(self):
        if self.norm == 1:
            return '{}({}->{})'.format(type(self).__name__, self.i, self.j)
        else:
            return '{} {}({}->{})'.format(self.norm, type(self).__name__, self.i, self.j)
    def __neg__(self):
        return type(self)(self.j, self.i, norm=self.norm, embedding=self.embedding)
    def __mul__(self, other):
        return type(self)(self.i, self.j, norm=self.norm*other, embedding=self.embedding)
    def __rmul__(self, other):
        return self.__mul__(other)
    def __add__(self, other):
        if isinstance(other, BondVectorSum):
            return other + self
        else:
            return BondVectorSum(self, other)
    def __sub__(self, other):
        return self + (-other)
    def __eq__(self, other):
        if not isinstance(other, BondVector):
            return self.dot(other) == 1
        else:
            return other.i == self.i and other.j == self.j
    def as_expr(self):
        return self.norm * (
                AnalyticModelBase.symbolic_x(self.j) - AnalyticModelBase.symbolic_x(self.i)
        ) / AnalyticModelBase.symbolic_r(self.j, self.i)
    @classmethod
    def get_embedding_angle_cos(self, i, j, k, l):
        # l is attached to k by definition
        return -sym.sin(AnalyticModelBase.symbolic_a(j, k, l)) * sym.sin(AnalyticModelBase.symbolic_t(i, j, k, l))
    @classmethod
    def get_positioning_cos(cls, i, j, k, l):
        # l is attached to k by definition
        a_ijk = AnalyticModelBase.symbolic_a(i,j,k)
        a_jkl = AnalyticModelBase.symbolic_a(j,k,l)
        t_ijkl = AnalyticModelBase.symbolic_t(i, j, k, l)
        cos = sym.cos; sin = sym.sin
        return cos(a_ijk)*cos(a_jkl) - sin(a_ijk)*sin(a_jkl)*cos(t_ijkl)
    def angle_cos(self, other, require_int=False):
        if isinstance(other, BondNormal):
            return other.angle_cos(self, require_int=require_int)
        elif isinstance(other, BondVectorSum):
            return other.distribute(lambda x:self.angle_cos(x, require_int=require_int))
        else:
            match_table = (
                self.i == other.i,
                self.i == other.j,
                self.j == other.i,
                self.j == other.j
            )
            match_sum = np.sum(match_table)
            if match_sum == 2: # two shared points
                if match_table[0]:
                    cos = 1
                else:
                    cos = -1
            elif match_sum == 0:
                return self.get_positioning_cos(self.i, self.j, other.i, other.j)
                # i = self.i
                # j = self.j
                # k = other.i
                # l = other.j
                #
                #
                # embedding = self.embedding
                # if embedding is None:
                #     embedding = other.embedding
                # if embedding is None:
                embedding = (self.i, self.j, other.i)
                self.logger.log_print(f"using embedding {embedding}")
                my_embedding = self.polar_components(embedding, self)
                other_embedding = other.polar_components(embedding, other)
                self.logger.log_print(f"emb1: {my_embedding}")
                self.logger.log_print(f"emb2: {other_embedding}")
                cos = my_embedding.dot(other_embedding).expand().simplify()
            else:
                if require_int:
                    return 2 + np.ravel_multi_index(match_table, (2, 2, 2, 2)) # binary encoding
                    # we'll build a composite binary integer so we can keep a single return type
                else:
                    if match_table[0]: # shared first atom
                        i = self.j
                        j = self.i
                        k = other.j
                        sign = 1
                    elif match_table[1]: # shared first and second
                        i = self.j
                        j = self.i
                        k = other.i
                        sign = -1
                    elif match_table[2]: # shared second and first
                        i = self.i
                        j = self.j
                        k = other.j
                        sign = -1
                    else: # shared second atom
                        i = self.i
                        j = self.j
                        k = other.i
                        sign = 1
                    cos = sign * sym.cos(AnalyticModelBase.symbolic_a(i, j, k))
            return cos
    def dot(self, other:'BondVector'):
        if isinstance(other, BondVectorSum):
            return other.distribute(
                lambda x:self.dot(x)
            ).simplify()
        else:
            with self.logger.block():
                self.logger.log_print("{s}", s=self)
                self.logger.log_print("{s}", s=other)
                acos = self.angle_cos(other)
                self.logger.log_print("acos: {acos}", acos=acos)
            return acos * self.norm * other.norm
    def cross(self, other):
        if isinstance(other, BondVectorSum):
            return other.distribute(
                lambda x:self.cross(x)
            )
        else:
            sangle = sym.sin(sym.acos(self.angle_cos(other))).trigsimp().subs(sym.Abs, sym.Id)
            return BondNormal(self, other,
                              norm=self.norm * other.norm * sangle,
                              embedding=self.embedding if self.embedding is not None else other.embedding
                              )

    @classmethod
    def polar_components(self, embedding_atoms, vector):
        raise NotImplementedError("dead code path")
        # returns the components and the basis for the embdedding
        # of atom l with respect to reference atoms i, j, and k
        i, j, k = embedding_atoms#sorted(embedding_atoms)
        basis = BondNormal.embedding_basis((i, j, k))
        # x-axis: i->j
        # z-axis: i->j X j->k
        n = vector.i
        m = vector.j
        sign = 1
        # We have two cases:
        #  1. all stuff inside i,j,k (simple rotations)
        #  2. i->l, j->l, or k->l (rotation + out of plane transform)
        matches = [n in (i, j, k), m in (i, j, k)]
        match_sum = np.sum(matches)

        if match_sum == 2: # simple rotation of main
            if m == j: # pointing towards center, should be flipped
                n, m = m, n
                sign *= -1
            if n == j:
                if m == i: # j->i is a flip of the basis vector
                    sign *= -1
                    angle = None
                else: # j->k, need i->j into j->k
                    angle = AnalyticModelBase.symbolic_a(i, j, k)
            else:
                if n == k: sign *= - 1
                # vector is i->k, need to rotation i->j into i->k
                angle = AnalyticModelBase.symbolic_a(j, i, k)

            if angle is None:
                components = OrientationVector(1, 0, 0, basis)
            else:
                self.logger.log_print("{n}->{m} in ({i},{j},{k}) = {a}", n=n,m=m,i=i,j=j,k=k,a=angle)
                rmat = OrientationVector.rotation_matrix(angle, OrientationVector(0, 0, 1, basis))
                v = OrientationVector(1, 0, 0, basis)
                components = rmat.dot(v)
        elif match_sum == 0:
            self.logger.log_print(f"{n}->{m} in ({i},{j},{k})")
            raise NotImplementedError("this shouldn't happen...")
        else:
            # rotation in plane and rotation out of plane
            if matches[1]: # want first point shared with embedding
                n, m = m, n
                sign *= -1

            if n == i: # simple dihedral
                # m<-i->j->k
                chob = None
                angle = AnalyticModelBase.symbolic_a(m, i, j)
                polar = AnalyticModelBase.symbolic_t(k, j, i, m)
            elif n == j:
                # i->j(->m)->k
                chob = None
                angle = AnalyticModelBase.symbolic_a(i, j, m)
                polar = AnalyticModelBase.symbolic_t(k, i, j, m)
            else:
                sign *= -1
                # need to rotate i->j->k->m so that ij -> kj
                chob = AnalyticModelBase.symbolic_a(i, j, k)
                angle = AnalyticModelBase.symbolic_a(j, k, m)
                polar = AnalyticModelBase.symbolic_t(i, j, k, m)

            self.logger.log_print("{n}->{m} in ({i},{j},{k}) = {a} | {p}", n=n,m=m,i=i,j=j,k=k,a=angle, p=polar)

            ax = OrientationVector(1, 0, 0, basis)
            v = ax
            if chob is not None:
                cr = OrientationVector.rotation_matrix(chob, OrientationVector(0, 0, 1, basis))
                v = cr.dot(v)
            amat = OrientationVector.rotation_matrix(angle, OrientationVector(0, 0, 1, basis))
            v = amat.dot(v)
            tmat = OrientationVector.rotation_matrix(polar, ax)
            v = tmat.dot(v)
            components = v
        self.logger.log_print(f"{n}->{m} in ({i},{j},{k}) -> {sign}")
        if sign < 0:
            components = -components
        return components
class BondNormal:
    def __init__(self, a, b, norm=1, embedding=None):
        self.a = a
        self.b = b
        if abs(self.a.angle_cos(self.b, require_int=True)) == 1:
            raise NotImplementedError('cross product of vector with itself ill defined')
        self.norm = norm
        if embedding is None:
            embedding = a.embedding
        if embedding is None:
            embedding = b.embedding
        if embedding is None:
            if isinstance(a, BondVector):
                i = a.i
                j = a.j
                k = b.i if b.i not in (i, j) else b.j
                embedding = (i, j, k)
        self.embedding = embedding
    @property
    def direction(self):
        return type(self)(self.b, self.a, norm=1, embedding=self.embedding)
    def __repr__(self):
        if self.norm == 1:
            return '{}({}x{})'.format(type(self).__name__, self.a, self.b)
        else:
            return '{} {}({}x{})'.format(self.norm, type(self).__name__, self.a, self.b)
    def __neg__(self):
        return type(self)(self.b, self.a, norm=self.norm, embedding=self.embedding)
    def __rmul__(self, other):
        return self.__mul__(other)
    def __mul__(self, other):
        return type(self)(self.a, self.b, norm=self.norm*other, embedding=self.embedding)
    def __add__(self, other):
        if isinstance(other, BondVectorSum):
            return other + self
        else:
            return BondVectorSum(self, other)
    def __sub__(self, other):
        if isinstance(other, BondVectorSum):
            return self + (-other)
        else:
            return BondVectorSum(self, -other)
    def __eq__(self, other):
        if not isinstance(other, BondNormal):
            return self.dot(other) == 1
        else:
            return self.a == other.a and self.b == other.b
    @classmethod
    def embedding_basis(cls, embedding):
        ei = BondVector(embedding[0], embedding[1])
        x = BondVector(embedding[1], embedding[2])
        ej = BondNormal(ei, x)
        ek = BondNormal(ei, ej)
        return [ei, ek, ej]
    @classmethod
    def polar_components(cls, embedding, self):
        raise NotImplementedError("dead code path")
        a = self.a.polar_components(embedding, self.a)
        a1, a2, a3 = a.vec
        b = self.b.polar_components(embedding, self.b)
        b1, b2, b3 = b.vec
        ang = sym.acos(self.a.angle_cos(self.b))
        sangle = sym.sin(ang).trigsimp().subs(sym.Abs, sym.Id)
        return OrientationVector( # gotta normalize...
            (a2*b3 - b2*a3)/sangle,
            (b1*a3 - a1*b3)/sangle,
            (a1*b2 - b1*a2)/sangle,
            a.basis
        )
    def angle_cos(self, other, require_int=True):
        if isinstance(other, BondVectorSum):
            return other.distribute(lambda v:self.angle_cos(v))
        if isinstance(other, BondVector) and isinstance(self.a, BondVector) and isinstance(self.b, BondVector):
            alignment_1 = self.a.angle_cos(other, require_int=True)
            alignment_2 = self.b.angle_cos(other, require_int=True)
            if abs(alignment_1) == 1 or abs(alignment_2) == 1: # cross product is perpendicular to contained vector
                cos = 0
            else:
                i = self.a.unshared_atom(self.b)
                j = self.a.shared_atom(self.b)
                k = self.b.unshared_atom(self.a)

                if other.i == i:
                    l = other.j
                    i,j,k = k,j,i
                    sign = -1
                elif other.j == i:
                    l = other.i
                    i,j,k = k,j,i
                    sign = 1
                elif other.i == j:
                    l = other.j
                    i,j,k = i,k,j
                    sign = -1
                elif other.j == j:
                    l = other.i
                    i,j,k = i,k,j
                    sign = 1
                elif other.i == k:
                    l = other.j
                    sign = 1
                elif other.j == k:
                    l = other.i
                    sign = -1
                else:
                    raise ValueError("5 atom angles unsupported")

                return sign * BondVector.get_embedding_angle_cos(i, j, k, l)

                # embedding = self.embedding
                # # if embedding is None:
                # #     embedding = other.embedding
                # if embedding is not None:
                #     my_embedding = self.polar_components(embedding, self)
                #     other_embedding = other.polar_components(embedding, other)
                #     BondVector.logger.log_print(f"emb1: {my_embedding}")
                #     BondVector.logger.log_print(f"emb2: {other_embedding}")
                #     cos = my_embedding.dot(other_embedding)
                # else:
                #     raise NotImplementedError("{}".format(self))
        else:
            # (axb).(cxd) = (a.c)(b.d) - (a.d)(b.c)
            #             = acos(a,c)acos(b,d) - acos(a,d)acos(b,c)
            aa = self.a.angle_cos(other.a)
            bb = self.b.angle_cos(other.b)
            ab = self.a.angle_cos(other.b)
            ba = self.b.angle_cos(other.a)
            # BondVector.logger.log_print(f"??? aa{aa} bb{bb} ab{ab} ba{ba}")

            ang_ab = sym.acos(self.a.angle_cos(self.b))
            sangle_ab = sym.sin(ang_ab).trigsimp().subs(sym.Abs, sym.Id)
            ang = sym.acos(other.a.angle_cos(other.b))
            sangle_cd = sym.sin(ang).trigsimp().subs(sym.Abs, sym.Id)
            return (aa*bb - ab*ba) / (sangle_ab*sangle_cd)

            alignments = [
                self.a.angle_cos(other.a, require_int=True),
                self.a.angle_cos(other.b, require_int=True),
                self.b.angle_cos(other.a, require_int=True),
                self.b.angle_cos(other.b, require_int=True)
            ]
            matches = [
                abs(x) == 1 for x in alignments
            ]
            match_sum = sum(matches)
            if match_sum == 0:
                raise NotImplementedError('angle between vector cross over different atoms not symbolically useful')
            elif match_sum == 2:
                # just need to determine the sign of alignment...
                if matches[0]: # same vectors basically, but we still need to account for signs
                    cos = alignments[0] * alignments[3]
                else: # flipped ordering so we pick up a negative sign
                    cos = -alignments[1] * alignments[2]
            else:
                # only one of the vectors matched, which means either we have some kind of dihedral angle
                # or we have zero, which we can check by checking to total cross product alignment
                if matches[0] or matches[2]:
                    subalignment = self.angle_cos(other.b, require_int=True)
                else:
                    subalignment = self.angle_cos(other.a, require_int=True)

                if abs(subalignment) == 1: # gotta be perpendicular...
                    cos = 0
                else: # need to pull out components to return the correct dihedral...
                    if matches[0]: # shared first bond, making these atoms j and k
                        i = self.b.unshared_atom(self.a)
                        j = self.b.shared_atom(self.a)
                        k = self.a.unshared_atom(self.b)
                        l = other.b.unshared_atom(other.a)
                        sign = alignments[0]
                    elif matches[1]:  # shared first and second
                        i = self.b.unshared_atom(self.a)
                        j = self.b.shared_atom(self.a)
                        k = self.a.unshared_atom(self.b)
                        l = other.a.unshared_atom(other.b)
                        sign = alignments[1]
                    elif matches[2]:  # shared second and first
                        i = self.a.unshared_atom(self.b)
                        j = self.a.shared_atom(self.b)
                        k = self.b.unshared_atom(self.a)
                        l = other.b.unshared_atom(other.a)
                        sign = alignments[2]
                    else: # shared second atom
                        i = self.a.unshared_atom(self.b)
                        j = self.a.shared_atom(self.b)
                        k = self.b.unshared_atom(self.a)
                        l = other.a.unshared_atom(other.b)
                        sign = alignments[3]
                    cos = sign * sym.cos(AnalyticModelBase.symbolic_t(i, j, k, l))

        return cos
    def as_expr(self):
        i = self.a.unshared_atom(self.b)
        j = self.a.shared_atom(self.b)
        k = self.b.unshared_atom(self.a)
        return self.norm * self.a.norm * self.b.norm * AnalyticModelBase.symbolic_n(k, j, i)
    def cross(self, other):
        if isinstance(other, BondVectorSum):
            return other.distribute(
                lambda x: self.cross(x)
            )
        else:
            sangle = sym.sin(sym.acos(self.angle_cos(other))).trigsimp().subs(sym.Abs, sym.Id)
            return BondNormal(self, other,
                              norm=self.norm * other.norm * sangle,
                              embedding=self.embedding if self.embedding is not None else other.embedding
            )
    def dot(self, other):
        if isinstance(other, BondVectorSum):
            return other.distribute(lambda v:self.dot(v))
        else:
            return self.angle_cos(other) * self.norm * other.norm
class BondVectorSum:
    def __init__(self, *terms):
        self.terms = terms
    def __add__(self, other):
        if isinstance(other, BondVectorSum):
            return type(self)(*self.terms, *other.terms)
        else:
            return type(self)(*self.terms, other)
    def __radd__(self, other):
        return self + other
    def __rmul__(self, other):
        return self.distribute(lambda a:other*a)
    def __mul__(self, other):
        return self.distribute(lambda a:other*a)
    def __repr__(self):
        return "{}({})".format(type(self).__name__, self.terms)
    def __neg__(self):
        return type(self)(*(-t for t in self.terms))
    def simplify(self):
        return sum(self.terms)

    def as_expr(self):
        return sum(
            b.as_expr() if hasattr(b, 'as_expr') else b
            for b in self.terms
        )
    def distribute(self, f):
        all_terms = [f(a) for a in self.terms]
        new_terms = ()
        for t in all_terms:
            new_terms = new_terms + (t.terms if isinstance(t, BondVectorSum) else (t,))
        return type(self)(*new_terms)
    def dot(self, other):
        return self.distribute(lambda a:a.dot(other))

class InternalJacobianDisplacements:

    @staticmethod
    def symbolic_m(i):
        return sym.Symbol('m[{i}]'.format(i=i), positive=True)

    @staticmethod
    def symbolic_r(i, j):
        if j < i:
            i, j = j, i
        return sym.Symbol('r[{i},{j}]'.format(i=i, j=j), positive=True)

    @staticmethod
    def symbolic_a(i, j, k):
        if k < i:
            i, k = k, i
        return sym.Symbol('a[{i},{j},{k}]'.format(i=i, j=j, k=k), real=True)

    @staticmethod
    def symbolic_t(i, j, k, l):
        if l < i:
            i, j, k, l = l, k, j, i
        return sym.Symbol('t[{i},{j},{k},{l}]'.format(i=i, j=j, k=k, l=l), real=True)

    @staticmethod
    def symbolic_y(i, j, k, l):
        return sym.Symbol('y[{i},{j},{k},{l}]'.format(i=i, j=j, k=k, l=l), real=True)

    @classmethod
    def lam(cls, i, j, k):
        r12 = cls.symbolic_r(i, j)
        r23 = cls.symbolic_r(j, k)
        a = cls.symbolic_a(i, j, k)
        return 1 / sym.sin(a) * (1 / r12 - sym.cos(a) / r23)

    @classmethod
    def dr(cls, i, j):
        return [
            BondVector(j, i),
            BondVector(i, j)
        ]

    @classmethod
    def da(cls, i, j, k):
        e12 = BondVector(i, j, embedding=(i, j, k))
        r12 = cls.symbolic_r(i, j)
        e23 = BondVector(j, k, embedding=(i, j, k))
        r23 = cls.symbolic_r(j, k)
        a = cls.symbolic_a(i, j, k)
        l321 = cls.lam(k, j, i)
        l123 = cls.lam(i, j, k)
        return [
            -1 / (r12 * sym.sin(a)) * (sym.cos(a) * e12 + e23),
            -l321 * e12 + l123 * e23,
            1 / (r23 * sym.sin(a)) * (e12 + sym.cos(a) * e23)
        ]

    @classmethod
    def dt(cls, i, j, k, l):
        e12 = BondVector(i, j, embedding=(i, j, k))
        r12 = cls.symbolic_r(i, j)
        e23 = BondVector(j, k, embedding=(i, j, k))
        r23 = cls.symbolic_r(j, k)
        e34 = BondVector(k, l, embedding=(i, j, k))
        r34 = cls.symbolic_r(k, l)
        a1 = cls.symbolic_a(i, j, k)
        a2 = cls.symbolic_a(j, k, l)
        l123 = cls.lam(i, j, k)
        l432 = cls.lam(l, k, j)
        x123 = BondNormal(e12, e23, embedding=(i, j, k))
        x234 = BondNormal(e23, e34, embedding=(j, k, l))
        return [
           -1 / (r12 * sym.sin(a1)) * x123,
            l123 * x123 - sym.cot(a2) / r23 * x234,
           -l432 * x234 + sym.cot(a1) / r23 * x123,
            1 / (r34 * sym.sin(a2)) * x234
        ]

    @classmethod
    def dy(cls, i, j, k, l):
        raise NotImplementedError("oops")

    @classmethod
    def displacement_vectors(cls, inds, coord_type=None):
        if coord_type is None:
            if len(inds) == 2:
                return cls.dr(*inds)
            elif len(inds) == 3:
                return cls.da(*inds)
            elif len(inds) == 4:
                return cls.dt(*inds)
            else:
                return cls.dy(*inds)
        else:
            return {
                "r":cls.dr,
                "a":cls.da,
                "t":cls.dt,
                "y":cls.dy
            }[coord_type](*inds)



import numpy as np

__all__ = [
    "PerturbationTheoryException",
    "Settings"
]
class PerturbationTheoryException(Exception):
    pass

class Settings:
    non_zero_cutoff = 1.0e-14

def _safe_dot(a, b):
    # generalizes the dot product so that we can use 0 as a special value...
    if (
            isinstance(a, (int, np.integer, float, np.floating)) and a == 0
            or isinstance(b, (int, np.integer, float, np.floating)) and b == 0
    ):
        return 0

    if isinstance(a, np.ndarray):
        doots = np.dot(a, b)
    else:
        doots = a.dot(b)

    if isinstance(b, np.ndarray) and isinstance(doots, SparseArray):
        doots = doots.asarray()

    return doots




__all__ = []
from .AnalyticModelConstructors import *; from .AnalyticModelConstructors import __all__ as exposed
__all__ += exposed
from .Helpers import *; from .Helpers import __all__ as exposed
__all__ += exposed


import enum
import functools

import numpy as np, scipy.linalg as slag, itertools, collections
import scipy.linalg

from McUtils.Data import AtomData, UnitsData
import McUtils.Numputils as nput

from .MixtureModes import MixtureModes

__all__ = [
    "NormalModes",
    "ReactionPathModes"
]

__reload_hook__ = [".MixtureModes"]

class NormalModes(MixtureModes):
    name="NormalModes"

    def __init__(self,
                 basis,
                 coeffs,
                 freqs=None,
                 origin=None,
                 masses=None,
                 inverse=None,
                 name=None,
                 mass_weighted=False,
                 frequency_scaled=False,
                 g_matrix=None
                 ):
        super().__init__(
            basis,
            coeffs,
            freqs=freqs,
            origin=origin,
            masses=masses,
            inverse=inverse,
            mass_weighted=mass_weighted,
            frequency_scaled=frequency_scaled,
            g_matrix=g_matrix,
            name=name
        )

    ModeData = collections.namedtuple("ModeData", ['freqs', 'modes', 'inverse'])
    default_zero_freq_cutoff = 1.0e-4 # 20 wavenumbers...
    @classmethod
    def get_normal_modes(cls,
                         f_matrix,
                         mass_spec,
                         # mass_units="AtomicMassUnits",
                         remove_transrot=True,
                         dimensionless=False,
                         mass_weighted=None,
                         zero_freq_cutoff=None,
                         return_gmatrix=False
                         ):

        f_matrix = np.asanyarray(f_matrix)
        if isinstance(mass_spec[0], str):  # atoms were supplied
            mass_spec = np.array([AtomData[a, "Mass"] for a in mass_spec]) * UnitsData.convert(
                "AtomicMassUnits",
                "AtomicUnitOfMass"
            )
        mass_spec = np.asanyarray(mass_spec)

        if mass_spec.ndim == 1:
            if mass_weighted is None: mass_weighted = dimensionless # generally do the right thing for Cartesians
            mass_spec = np.broadcast_to(mass_spec[:, np.newaxis], (len(mass_spec), 3)).flatten()
            mass_spec = np.diag(1 / mass_spec)
        freq2, modes = slag.eigh(f_matrix, mass_spec, type=3)

        gi12 = nput.fractional_power(mass_spec, -1 / 2)
        g12 = nput.fractional_power(mass_spec, 1 / 2)
        V = gi12 @ modes
        # modes = g12 @ V
        # therefore, inv = V.T @ gi12
        # and modes[:, nz] = g12 @ V[:, nz]; inv[nz, :] = (V.T)[nz, :] @ gi12 (this is only the _left_ inverse)

        if remove_transrot:
            if zero_freq_cutoff is None:
                zero_freq_cutoff = cls.default_zero_freq_cutoff
            nonzero = np.abs(freq2) > zero_freq_cutoff**2
        # if len(nonzero) < len(modes):
        #     if np.linalg.det(modes) != 1 or not np.allclose(modes.T @ modes, np.eyelen(modes)):
        #         # we're working in a non-invertible subspace...
        #         raise ValueError("non-invertible subspace of normal modes found, inverse can't be evaluated")
        #     else:
        #         inv = modes.T
        # else:
        #     inv = np.linalg.inv(modes)
        # nonzero = np.array([0, 3, 4, 5])
            freq2 = freq2[nonzero]
            modes = modes[:, nonzero]
            inv = V.T[nonzero, :] @ gi12
        else:
            inv = V.T @ gi12
        # else:
        #     if np.linalg.det(modes) == 1 and np.allclose(modes.T @ modes, np.eyelen(modes)):
        #         inv = modes.T
        #     else:
        #         inv = np.linalg.inv(modes)

        freqs = np.sign(freq2) * np.sqrt(np.abs(freq2))
        if mass_weighted:
            # gi12 = slag.fractional_matrix_power(mass_spec, -1 / 2)
            # g12 = slag.fractional_matrix_power(mass_spec, 1 / 2)
            inv = inv @ g12
            modes = gi12 @ modes
        if dimensionless:
            weighting = np.sqrt(np.abs(freqs))
            modes = modes / weighting[np.newaxis, :]
            inv = inv * weighting[:, np.newaxis]
        # if mode == 'reasonable':
        # modes, inv = inv.T, modes.T

        sorting = np.argsort(freqs)
        modes = modes[:, sorting]
        inv = inv[sorting, :]

        mode_data = cls.ModeData(freqs, inv.T, modes.T)
        if return_gmatrix:
            return mode_data, mass_spec
        else:
            return mode_data

    @classmethod
    def from_fg(cls,
                basis,
                f_matrix,
                mass_spec,
                remove_transrot=True,
                dimensionless=False,
                zero_freq_cutoff=None,
                mass_weighted=None,
                origin=None,
                **opts
                ):
        """
        Generates normal modes from the specified F and G matrices

        :param basis:
        :param f_matrix: second derivatives of the potential
        :param mass_spec:
        :param mass_units:
        :param remove_transrot:
        :param opts:
        :return:
        """

        (freqs, modes, inv), g_matrix = cls.get_normal_modes(f_matrix, mass_spec,
                                                 remove_transrot=remove_transrot,
                                                 dimensionless=dimensionless,
                                                 zero_freq_cutoff=zero_freq_cutoff,
                                                 mass_weighted=mass_weighted,
                                                 return_gmatrix=True
                                                 )

        mass_weighted = mass_weighted or dimensionless

        if mass_weighted and origin is not None:
            origin = np.asanyarray(origin)
            origin = (origin.flatten() @ nput.fractional_power(g_matrix, -1/2)).reshape(origin.shape)

        return cls(basis, modes, inverse=inv, freqs=freqs,
                   mass_weighted=mass_weighted,
                   frequency_scaled=dimensionless,
                   g_matrix=g_matrix,
                   origin=origin,
                   **opts
                   )

    default_projected_zero_freq_cutoff = None
    @classmethod
    def from_molecule(cls, mol,
                      dimensionless=False, use_internals=None,
                      project_transrot=True,
                      zero_freq_cutoff=None,
                      masses=None,
                      **opts
                      ):
        from ..Molecools import Molecule

        mol = mol  # type:Molecule
        if use_internals is None:
            use_internals = mol.internal_coordinates is not None
        if use_internals:
            hess = nput.tensor_reexpand(
                mol.get_cartesians_by_internals(1, strip_embedding=True, reembed=True),
                [0, mol.potential_derivatives[1]],
                order=2
            )[1]
            if zero_freq_cutoff is None:
                zero_freq_cutoff = cls.default_projected_zero_freq_cutoff
                if zero_freq_cutoff is None:
                    zero_freq_cutoff = 1 * UnitsData.convert("Wavenumbers", "Hartrees")
            return cls.from_fg(
                mol.internal_coordinates.system,
                hess,
                mol.get_gmatrix(masses=masses),
                dimensionless=dimensionless,
                origin=mol.get_internals(strip_embedding=True),
                zero_freq_cutoff=zero_freq_cutoff,
                **opts
            )
        else:
            if masses is None:
                masses = mol.atomic_masses
            hess = mol.potential_derivatives[1]
            if project_transrot:
                proj = mol.get_translation_rotation_projector(mass_weighted=True)
                mw = np.diag(np.repeat(1 / np.sqrt(masses), 3))
                mw_inv = np.diag(np.repeat(np.sqrt(masses), 3))
                proj = mw_inv @ proj @ mw
                hess = proj @ hess @ proj.T
                if zero_freq_cutoff is None:
                    zero_freq_cutoff = cls.default_projected_zero_freq_cutoff
                    if zero_freq_cutoff is None:
                        zero_freq_cutoff = 1 * UnitsData.convert("Wavenumbers", "Hartrees")

            return cls.from_fg(
                mol.coords.system,
                hess,
                masses,
                dimensionless=dimensionless,
                masses=masses,
                origin=mol.coords,
                zero_freq_cutoff=zero_freq_cutoff,
                **opts
            )

    @classmethod
    def _atom_projector(cls, n, i):
        if nput.is_numeric(i):
            i = [i]
        z = np.zeros((3 * n, 3 * n))
        for i in i:
            x = np.arange(3 * i, 3 * (i + 1))
            z[x, x] = 1
        return z
    def get_nearest_mode_transform(self,
                                   alternate_modes:np.ndarray,
                                   mass_weighted=False,
                                   atoms=None,
                                   maximum_similarity=True,
                                   unitarize=True
                                   ):
        if not mass_weighted:
            modes = self.remove_mass_weighting()
        else:
            modes = self.make_mass_weighted()

        inv = modes.inverse
        modes = modes.matrix
        if atoms is not None:
            proj = self._atom_projector(modes.shape[0], atoms)
            modes = proj @ modes
            alternate_modes = proj @ alternate_modes

        if maximum_similarity:
            tf = nput.maximum_similarity_transformation(modes, alternate_modes, apply_transformation=False)
            return tf, tf.T
        else:
            tf = inv @ alternate_modes
            if unitarize:
                tf = nput.unitarize_transformation(tf)
                inv = tf.T
            else:
                inv = np.linalg.pinv(tf)
            return tf, inv

    def get_atom_localized_mode_transformation(self,
                                               atoms,
                                               masses=None, origin=None,
                                               localization_type='ned',
                                               allow_mode_mixing=False,
                                               maximum_similarity=False,
                                               unitarize=True
                                               ):
        if masses is None:
            masses = self.masses
            mw = self.make_mass_weighted()
        else:
            mw = self.make_mass_weighted(masses=masses)

        if origin is None:
            origin = mw.remove_mass_weighting().origin

        if nput.is_numeric(atoms):
            atoms = [atoms]

        f = mw.matrix @ np.diag(self.freqs**2) @ mw.matrix.T

        if localization_type == 'direct':
            tr_tf, tr_inv = nput.translation_rotation_invariant_transformation(
                origin.reshape(-1, 3),
                masses,
                mass_weighted=True
            )
        else:
            tr_tf = tr_inv = np.eye(mw.matrix.shape[0])

        g_matrix = tr_tf.T @ tr_tf
        if allow_mode_mixing:
            proj = self._atom_projector(len(masses), atoms)
            freqs, modes, inv = self.get_normal_modes(
                tr_inv @ proj @ f @ proj @ tr_inv.T,
                g_matrix,
                remove_transrot=True
            )
        else:
            # freqs = []
            modes = []
            # inv = []
            for atom in atoms:
                proj = self._atom_projector(len(masses), atom)
                sub_freqs, sub_modes, sub_inv = self.get_normal_modes(
                    tr_inv @ proj @ f @ proj @ tr_inv.T,
                    g_matrix,
                    remove_transrot=True
                )
                # freqs.append(sub_freqs)
                modes.append(sub_modes)
                # inv.append(sub_inv)
            # freqs = np.concatenate(freqs)
            modes = np.concatenate(modes, axis=1)
            # inv = np.concatenate(inv, axis=0)

        if localization_type == 'direct':
            modes = tr_tf @ modes

        return mw.get_nearest_mode_transform(
            modes,
            mass_weighted=True,
            maximum_similarity=maximum_similarity,
            unitarize=unitarize
        )

    def get_internal_localized_mode_transformation(
            self,
            expansion_coordinates: "Iterable[Iterable[int]|dict]",
            fixed_atoms=None,
            mass_weighted=False,
            project_transrot=True,
            atoms=None,
            maximum_similarity=False,
            unitarize=True
    ):
        # from .ObliqueModes import ObliqueModeGenerator

        nmw = self.remove_mass_weighting()
        base_derivs = nput.internal_coordinate_tensors(
            nmw.origin,
            expansion_coordinates,
            fixed_atoms=fixed_atoms
        )

        if mass_weighted:
            base_derivs = np.diag(np.repeat(1/np.sqrt(self.masses), 3)) @ base_derivs

        if project_transrot:
            origin = self.origin
            masses = self.masses
            projector = nput.translation_rotation_projector(
                    origin.reshape(-1, 3),
                    masses,
                    mass_weighted=mass_weighted
                )
            base_derivs = projector @ base_derivs

        return self.get_nearest_mode_transform(
            base_derivs,
            atoms=atoms,
            mass_weighted=mass_weighted,
            maximum_similarity=maximum_similarity,
            unitarize=unitarize
        )

    def get_displacement_localized_mode_transformation(self,
                                                       mode_blocks=None,
                                                       atoms=None,
                                                       mass_weighted=True,
                                                       unitarize=True,
                                                       **maximizer_opts
                                                       ):
        if not unitarize:
            raise ValueError("Jacobi iterations only give a unitary transformation")

        if mass_weighted:
            modes = self.make_mass_weighted().matrix
        else:
            modes = self.remove_mass_weighting().matrix

        if atoms is not None:
            modes = self._atom_projector(modes.shape[0], atoms) @ modes
        if mode_blocks is None:
            mode_blocks = [np.arange(modes.shape[1])]
        elif nput.is_numeric(mode_blocks[0]):
            mode_blocks = [mode_blocks]

        tf_inv = np.zeros((modes.shape[1], modes.shape[1]), dtype=float)
        for block in mode_blocks:
            _, sub_inv, _ = nput.jacobi_maximize(
                modes[:, block],
                nput.displacement_localizing_rotation_generator,
                **maximizer_opts
            )
            tf_inv[np.ix_(block, block)] = sub_inv

        return tf_inv, tf_inv.T

    class LocalizationMethods(enum.Enum):
        MaximumSimilarity = 'target_modes'
        AtomLocalized = 'atoms'
        DisplacmentMinimized = 'displacements'
        Internals = 'coordinates'

    @property
    def localizer_dispatch(self):
        return {
            self.LocalizationMethods.MaximumSimilarity.value:(self.get_nearest_mode_transform, 'target_modes'),
            self.LocalizationMethods.AtomLocalized.value:(self.get_atom_localized_mode_transformation, 'atoms'),
            self.LocalizationMethods.DisplacmentMinimized.value:(self.get_displacement_localized_mode_transformation, 'mode_blocks'),
            self.LocalizationMethods.Internals.value:(self.get_internal_localized_mode_transformation, 'internals')
        }

    def localize(self,
                 method=None,
                 *,
                 atoms=None,
                 target_modes=None,
                 internals=None,
                 mode_blocks=None,
                 reorthogonalize=None,
                 unitarize=True,
                 **opts
                 ):

        from .LocalizedModes import LocalizedModes

        if method is None:
            if target_modes is not None:
                method = 'target_modes'
            elif internals is not None:
                method = 'coordinates'
            elif mode_blocks is not None:
                method = 'displacements'
            elif atoms is not None:
                method = 'atoms'
            else:
                method = 'displacements'

        if isinstance(method, str):
            method = self.LocalizationMethods(method)

        if hasattr(method, 'value'):
            method = self.localizer_dispatch.get(method.value, method)

        args = ()
        try:
            method, arg_names = method
        except TypeError:
            ...
        else:
            if isinstance(arg_names, str):
                arg_names = [arg_names]
            all_kw = dict(
                atoms=atoms,
                target_modes=target_modes,
                internals=internals,
                mode_blocks=mode_blocks,
                reorthogonalize=reorthogonalize,
                unitarize=unitarize
            )
            args = tuple(all_kw.get(k) for k in arg_names)
            if 'atoms' not in arg_names:
                opts['atoms'] = atoms # taken by all of the localizers

        tf, inv = method(*args, unitarize=unitarize, **opts)
        # inverse = tf.T @ self.inverse # assumes a unitary localization

        if reorthogonalize is None:
            reorthogonalize = not unitarize
        if reorthogonalize:
            modes = self.make_mass_weighted().matrix @ tf
            tf = tf @ nput.fractional_power(modes.T @ modes, -1 / 2)
            inv = nput.fractional_power(modes.T @ modes, 1 / 2) @ inv

        return LocalizedModes(self, tf, inverse=inv)

class ReactionPathModes(NormalModes):
    def get_rp_modes(cls,
                  gradient,
                  f_matrix,
                  mass_spec,
                  mass_units="AtomicMassUnits",
                  remove_transrot=True,
                  dimensionless=False
                  ):
        ...